/*
 *
 * fvwm/FvwmModule.cxx
 *
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include <unistd.h>
#include <stdio.h>

#include <X11/Xlib.h>

#include <nucleo/core/ReactiveEngine.H>

#include "wncdesktop/XwncDesktop.H"
#include "wncdesktop/XwncWindow.H"
#include "main/AScreen.H"

#include "FvwmModule.H"
#include "AFvwm.H"

#include <nucleo/utils/FileUtils.H>

void FvwmModule::_processOperation(XwncWindow *xwncWin, char *rest)
{
	char *op = NULL;
	int drag_rotate = DRAG_ROTATE_NO;

	rest = GetNextToken(rest, &op);

	int val[2];
	if (!rest || GetIntegerArguments(rest, &rest, val, 2) != 2)
	{
		/* should not happen */
		return;
	}

	if (StrEquals(op, "Fold"))
	{
		_ascr->fold(xwncWin, val[0], val[1]);
	}
	else if (StrEquals(op, "Move"))
	{
		float alpha = 0.7;

		if (rest)
		{
			char *a;
			a = PeekToken(rest, &rest);
			if (a && sscanf(a,"%f",&alpha))
			{
				if (alpha > 1.0)
					alpha = 1.0;
				else if (alpha < 0.1)
					alpha = 0.1;
			}
			else
			{
				alpha = 0.7;
			}
		}
		_ascr->startInternalMove(
			xwncWin, val[0], val[1], alpha);
	}
	else if (StrEquals(op,"DragRotateXY"))
	{
		drag_rotate = DRAG_ROTATE_XY;
	}
	else if (StrEquals(op,"DragRotateX")) {
		drag_rotate = DRAG_ROTATE_X;
	}
	else if (StrEquals(op,"DragRotateY")) {
		drag_rotate = DRAG_ROTATE_Y;
	}
	else if (StrEquals(op,"DragRotateZ")) {
		drag_rotate = DRAG_ROTATE_Z;
	}

	if (drag_rotate != DRAG_ROTATE_NO)
	{
		_ascr->dragRotate(xwncWin, drag_rotate);
	}

	if (op)
	{
		free(op);
	}
}

void FvwmModule::_parseMessageLine(Window id, char *line)
{
	char *opt;
	char *rest;

	XwncWindow *xwncWin = _wncDesktop->findWindow(id);

	AbstractWindowRenderer *winRenderer = 0;
	
	if (xwncWin)
	{
		winRenderer = xwncWin->getRenderer();
	}

	opt = PeekToken(line, &rest);
	if (StrEquals(opt,"Scale"))
	{
		if (winRenderer)
		{
			float pct;
			opt = PeekToken(rest, &rest);
			if (opt && sscanf(opt,"%f",&pct))
			{
				_ascr->root->scaleWindow(
					xwncWin, pct,pct,pct);
			}
		}
	}
	else if (StrEquals(opt,"AbsoluteScale"))
	{
		if (winRenderer)
		{
			float pct;
			opt = PeekToken(rest, &rest);
			if (opt && sscanf(opt,"%f",&pct))
			{
				_ascr->root->scaleWindow(
					xwncWin, pct,pct,pct, 0, true);
			}
		}
	}
	else if (StrEquals(opt,"Rotate"))
	{
		int val[4];
		if (GetIntegerArguments(rest, NULL, val, 4) == 4)
		{
			if (winRenderer)
			{
				_ascr->root->rotateWindow(
					xwncWin, val[0],val[1],val[2],val[3]);
				//winRenderer->rotate_rel(
				//val[0],val[1],val[2],val[3]);
			}
		}
	}
	else if (StrEquals(opt,"AbsoluteRotate"))
	{
		int val[4];
		if (GetIntegerArguments(rest, NULL, val, 4) == 4)
		{
			if (xwncWin)
			{
				//xwncWin->absoluteRotate(
				// val[0],val[1],val[2],val[3]);
				_ascr->root->rotateWindow(
					xwncWin, val[0],val[1],val[2],val[3],
					0, true);
			}
		}
	}
	else if (StrEquals(opt,"Reset"))
	{
		if (winRenderer)
		{
			winRenderer->resetTransformations();
			xwncWin->restoreTranslation();
		}
	}
	else if (StrEquals(opt,"Operation"))
	{
		/* fvwm send us a window */
		if (xwncWin)
		{
			_processOperation(xwncWin, rest);
		}
	}
	else if (StrEquals(opt,"AllowsSurfaceRotation"))
	{
		if (xwncWin)
		{
			xwncWin->surfaceRotation(false);
		}
	}
	else if (StrEquals(opt,"NoSurfaceRotation"))
	{
		if (xwncWin)
		{
			xwncWin->surfaceRotation(true);
		}
	}
	else if (StrEquals(opt,"Surface") || StrEquals(opt,"DefineSurface"))
	{
		int surface = -1;
		int toggle;
		float ft[4];
		int nbr = 0;

		if (StrEquals(opt,"Surface"))
		{
			toggle = ParseToggleArgument(rest, &rest, -1, 0);
			_ascr->root->enableSurface(toggle);
		}

		opt = PeekToken(rest, &rest);

		if (StrEquals(opt,"Sphere"))
		{
			surface = SURFACE_SPHERE;
		}
		else if (StrEquals(opt,"Cylinder") ||
			 StrEquals(opt,"LRCylinder"))
		{
			surface = SURFACE_LRCYLINDER;	
		}
		else if (StrEquals(opt,"TBCylinder"))
		{
			surface = SURFACE_TBCYLINDER;	
		}
		else if (StrEquals(opt,"SphereCylinder"))
		{
			surface = SURFACE_SPHERE_CYLINDER;	
		}
		else if (StrEquals(opt,"CylinderSphere"))
		{
			surface = SURFACE_CYLINDER_SPHERE;	
		}
		else if (StrEquals(opt,"Current"))
		{
			surface = -2;	
		}

		if (surface != -1)
		{
			opt = PeekToken(rest, &rest);
		}

		while(opt && *opt != '\0')
		{
			float f;

			if (nbr < 4 && sscanf(opt,"%f",&f))
			{
				ft[nbr++] = f;
			}
			opt = PeekToken(rest, &rest);
		}

		if (surface == -2)
		{
			if (nbr >= 2)
			{
				_ascr->root->setSurface(ft[0], ft[1]);
			}
			else if (nbr >= 1)
			{
				_ascr->root->setSurface(ft[0]);
			}
		}
		else if (surface != -1)
		{
			if (nbr >= 2)
			{
				_ascr->root->setSurface(surface, ft[0], ft[1]);
			}
			else if (nbr >= 1)
			{
				_ascr->root->setSurface(surface, ft[0]);
			}
			else
			{
				_ascr->root->setSurface(surface);
			}
		}
		else
		{
			if (nbr >= 4)
			{
				_ascr->root->setSurfaceParameters(
					ft[0], ft[1], ft[2], ft[3]);
			}
			else if (nbr >= 2)
			{
				_ascr->root->setSurfaceParameters(ft[0], ft[1]);
			}
			else if (nbr >= 1)
			{
				_ascr->root->setSurfaceParameters(ft[0]);
			}
		}
	}
	else if (StrEquals(opt,"CropToCircule"))
	{
		if (winRenderer)
		{
			winRenderer->cropToCircule() ;
		}
	}
	else if (StrEquals(opt,"TestPolygon"))
	{
		if (winRenderer)
		{
			winRenderer->testPolygon() ;
		}
	}
	else if (StrEquals(opt,"Animate"))
	{
		if (winRenderer)
		{
			winRenderer->animate() ;
		}
	}
	else if (StrEquals(opt,"Shadow"))
	{
		if (winRenderer)
		{
			winRenderer->shadow() ;
		}
	}
	else if (StrEquals(opt,"SaveFrame"))
	{
		_ascr->saveFrame();
	}
	else if (StrEquals(opt,"ShowWindow"))
	{
		if (xwncWin)
		  {
			_ascr->root->translateForWindow(xwncWin);
			_ascr->draw(false);
		}
	}
	else if (StrEquals(opt,"MoreTransparency"))
	{
		if (xwncWin)
		{
			float d = 0.05;
			float a,pct;
			a = xwncWin->getAlpha();
			opt = PeekToken(rest, &rest);
			if (opt && sscanf(opt,"%f",&pct))
			{
				if (pct > 1.0) pct = 1.0;
				else if (pct < 0) pct = 0;
				d = pct;
			}
			a -= d;
			if (a < 0.05) a = 0.05;
			xwncWin->setAlpha(a);
		}
	}
	else if (StrEquals(opt,"LessTransparency"))
	{
		if (xwncWin)
		{
			float d = 0.05;
			float a,pct;
			a = xwncWin->getAlpha();
			opt = PeekToken(rest, &rest);
			if (opt && sscanf(opt,"%f",&pct))
			{
				if (pct > 1.0) pct = 1.0;
				else if (pct < 0) pct = 0;
				d = pct;
			}
			a += d;
			if (a > 1.0) a = 1;
			xwncWin->setAlpha(a);
		}
	}
	else if (StrEquals(opt,"SetOrRemoveTransparency"))
	{
		if (xwncWin)
		{
			float pct;
			float a = xwncWin->getAlpha();
			opt = PeekToken(rest, &rest);
			if (opt && a == 1.0 && sscanf(opt,"%f",&pct))
			{
				if (pct > 1.0) pct = 1.0;
				else if (pct < 0) pct = 0.05;
				xwncWin->setAlpha(pct);
			}
			else
			{
				xwncWin->setAlpha(1.0);
			}
		}
	}
	else if (StrEquals(opt,"SetTransparency"))
	{
		if (xwncWin)
		{
			float pct;
			opt = PeekToken(rest, &rest);
			if (opt && sscanf(opt,"%f",&pct))
			{
				if (pct > 1.0) pct = 1.0;
				else if (pct < 0) pct = 0;
				xwncWin->setAlpha(pct);
			}
		}
	}
	else if (StrEquals(opt,"ResizeTransparency"))
	{
		float pct;
		opt = PeekToken(rest, &rest);
		if (opt && sscanf(opt,"%f",&pct))
		{
			if (pct > 1.0) pct = 1.0;
			else if (pct < 0) pct = 0;
			_ascr->setResizeTransparency(pct);
		}
	}
	else if (StrEquals(opt,"GlobalTransparency"))
	{
		int toggle = ParseToggleArgument(rest, NULL, -1, 0);
		if (toggle == -1) {
			_ascr->toggleTransparent();
		}
		else {
			_ascr->setTransparent((bool)toggle);
		}
	}
	else if (StrEquals(opt,"RootTranslate")) {
		int val[3];
		if (GetIntegerArguments(rest, NULL, val, 3) == 3)
		{
			_ascr->root->translate_rel(val[0],val[1],val[2]);
			_ascr->resetSelection();
		}
	}
	else if (StrEquals(opt,"RootRotate")) {
		int val[4];
		if (GetIntegerArguments(rest, NULL, val, 4) == 4)
		{
			_ascr->root->rotate(val[0],val[1],val[2],val[3]);
		}
	}
	else if (StrEquals(opt,"RootScale"))
	{
		float pct;
		opt = PeekToken(rest, &rest);
		if (opt && sscanf(opt,"%f",&pct))
		{
			_ascr->root->scale(pct,pct,1);
			_ascr->resetSelection();
		}
	}
	else if (StrEquals(opt,"RootReset"))
	{
		_ascr->root->resetTransformations();
		_ascr->resetSelection();
	}
	else if (StrEquals(opt,"ExposeMode"))
	{
		_ascr->root->exposeMode();
	}
	else if (StrEquals(opt,"Restart"))
	{
		sendText((char *)_restartString.c_str(), 0);
	}
	else if (StrEquals(opt,"Quit") || StrEquals(opt,"Exit"))
	{
		// ask fvwm to quit
		SendText(_fd, "Quit", 0);
	}
	else
	{
		std::cerr << "[" <<_myName << "] WARNING -- " <<
			"Unknown command " << line << std::endl;
	}
}

void
FvwmModule::_processMessage(FvwmPacket* packet)
{
	unsigned long type = packet->type;
	// unsigned long length = packet->size;
	unsigned long* body = packet->body;
	XwncWindow *win = 0;
	static int moveX, moveY;

	switch (type)
	{
	case M_STRING:
		_parseMessageLine((Window)body[1], (char *)&body[3]);
		break;
	case M_NEW_DESK:
		if (_currentDesk == (long)body[0])
		{
			SendUnlockNotification(_fd);
		}
		else
		{
			_currentDesk = (long)body[0];
			int step;
			int d = 20;
			for (step = 1; step <= 180; step += d)
			{
				_ascr->root->rotate(d,0,d,0);
				_ascr->draw(false);
				usleep(1000);
			}
			SendUnlockNotification(_fd);
			for (step = 1; step <= 180; step += d)
			{
				_ascr->root->rotate(d,0,d,0);
				_ascr->draw(false);
				usleep(1000);
			}
			_ascr->resetSelection();
		}
		break;
	case MX_STARTSTOP_OPERATION:
		/* Warning: fvwm wait for an unlock notification */
#if 0
		fprintf(stderr, "MX_STARTSTOP_OPERATION "
			"0x%lx, 0x%lx, %lu, %lu\n",
			(Window)body[1], (Window)body[0], body[2], body[3]);
#endif
		win = _wncDesktop->findWindow((Window)body[1]);
		if (win)
		{
			bool start = false;
			if (body[3] == MX_STARTSTOP_OPERATION_START)
			{
				start = true;
			}
			if (body[2] == MX_STARTSTOP_OPERATION_MOVE)
			{
				_ascr->wmMove(win, start);
			}
			else if (body[2] == MX_STARTSTOP_OPERATION_RESIZE)
			{
				_ascr->wmResize(win, start);
			}
			else if (body[2] == MX_STARTSTOP_OPERATION_MODULE_MOVE)
			{
				_ascr->wmMove(win, start);
				if (!start)
				{
					int x,y;
					char cmd[256];

					win->getRealPosition(&x, &y);
					sprintf(cmd, "Move w%ip w%ip",
						moveX - x, moveY - y);
					sendText(cmd, win);
				}
				else
				{
					win->getRealPosition(&moveX, &moveY);
				}
			}
		}
		/* unlock fvwm */
		SendUnlockNotification(_fd);
		break;
	case MX_MODULE_INTERACTIVE_MOVE:
	{
#if 0
		fprintf(stderr, "MX_MODULE_INTERACTIVE_MOVE "
			"0x%lx, 0x%lx, %i, %i\n",
			(Window)body[1], (Window)body[0],
			(int)body[2], (int)body[3]);
#endif
		int x = (int)body[2];
		int y = (int)body[3];
		win = _wncDesktop->findWindow((Window)body[1]);
		if (win)
		{
			int oldX, oldY;
			win->getRealPosition(&oldX, &oldY);

			if (x != oldX || y != oldY)
			{
				win->translate_rel(-oldX + x, oldY - y, 0);
				win->setRealPosition(x, y);
			}
		}
		SendUnlockNotification(_fd);
	}
	break;
#if 0
	case MX_CLICK_MOTION:
	{
		win = _wncDesktop->findWindow((Window)body[1]);
		unsigned long isDNDMotion = (unsigned long)body[2];
		if (win && !isDNDMotion)
		{
			_ascr->clickMotion(win);
		}
	}
	break;
#endif
	default:
		/* ignore unknown packet */
		break;
    }
}

void
FvwmModule::react(Observable *obs)
{
  if (obs==_fk) {
    FvwmPacket* packet = ReadFvwmPacket(_fd[1]);
    if ( packet == NULL )
	 {
	   // FIXME, a more clean exit. In general we are here when
	   // fvwm exit
	   exit(0);
	 }
    else
	 {
	   _processMessage(packet);
	 }
  }
  notifyObservers() ;
}

void
FvwmModule::sendText(char *text, XwncWindow *win)
{
	SendText(_fd, text, (win)? win->getFrameID() : 0);
}

FvwmModule::FvwmModule(AScreen *ascr, XwncDesktop *wncDesktop,
				   int fd0, int fd1,
				   char *myName,
				   std::string restartString)
{

	_myName = myName;
	_restartString = restartString ;
	_ascr = ascr;
	_wncDesktop = wncDesktop;

	_fd[0] = fd0;
	_fd[1] = fd1;

	_currentDesk = 0;

	/* the msg type we want from fvwm (see libs/Module.h and the fvwm
	 * www site for more msg and info) */
	SetMessageMask( _fd, M_STRING|M_NEW_DESK);
	/* extended messages */
	SetMessageMask(
		_fd,
		MX_STARTSTOP_OPERATION     |
		MX_MODULE_INTERACTIVE_MOVE |
		MX_CLICK_MOTION);

	/* should synchronouse for these msg */
	SetSyncMask(_fd, M_NEW_DESK);
	SetSyncMask(_fd, MX_STARTSTOP_OPERATION|MX_MODULE_INTERACTIVE_MOVE);

	_fk = FileKeeper::create(_fd[1], FileKeeper::R);
	subscribeTo(_fk);
	ReactiveEngine::notify(this, this); 

	_ascr->setFvwmModule(this);

	/* tell fvwm we're running */
	SendFinishedStartupNotification(_fd);
}

