/*
 *
 * main/LayerManaer.cxx --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

// for gravity bits (not really necessary)
#include "X11/X.h"

#include "wncdesktop/XwncWindow.H"
#include "LayerManager.H"

// --------------------------------------------------------------------
// stacking

void LayerManager::_getParameters(
	float *unmanagedDepth, float *managedDepth, int *unmanagedNum,
	int *managedNum)
{
	float ud = 0;
	float md = 0;
	int mn = 0;
	int un = 0;

	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
    
		if (!win)
		{
			/* should not happen */
			continue;
		}
		if (win->isRootWindow() || win->isEwmhDesktop())
		{
			continue;
		}

		if (win->isUnmanaged())
		{
			un++;
		}
		else
		{
			mn++;
		}
	}
	
	// compute a reasonable intermediate depths:
	if (_eyeDist == 0)
	{
		// otho proj
		// FIXME: should do a "layer correction" as for the "real" proj
		md = 0.1;
	}
	else if (mn > 0)
	{
		md = _bigGap/(2*mn);
	}
	else
	{
		// do not care
		md = 10;
	}
	
	if (_eyeDist == 0)
	{
		// otho proj
		// FIXME: should do a "layer correction" as for the "real" proj
		ud = 0.1;
	}
	else if (un > 0)
	{
		ud = _bigGap/(2*un);
	}
	else
	{
		// do not care
		ud = 10;
	}
	
	*managedNum = mn;
	*unmanagedNum = un;
	*managedDepth = md;
	*unmanagedDepth = ud;
}

float LayerManager::_getDeltaDepth(
	XwncWindow *win, int *count, float unmanagedDepth, float managedDepth,
	int numUnmanaged, int numManaged)
{
	if (win->isRootWindow())
	{
		// the root window
		return 0;
	}
	if (win->isEwmhDesktop())
	{
		// put it just after the root window
		return 1;
	}

	float d = 0;
	*count = *count + 1;

	if (*count <= numManaged)
	{
		// a managed window
		if (*count == 1)
		{
			// first managed
			d =+ _bigGap;
		}
		if (*count == numManaged)
		{
			// last managed
			d =+ _bigGap - ((numManaged-1) * managedDepth);
		}
		if (*count > 1 && *count < numManaged)
		{
			d =+ managedDepth;
		}
		return d;
	}

	// unmanaged window
	int c = *count - numManaged;
	if (c == 1)
	{
		// first unmanaged
		d =+ _bigGap;
	}
	if (c == numUnmanaged)
	{
		// last unmanaged
		d =+ _bigGap - ((numUnmanaged-1) * unmanagedDepth);
	}
	if (c > 1 && c < numUnmanaged)
	{
		d =+ unmanagedDepth;
	}
	return d;
}
 
void LayerManager::display(dlPolicy policy)
{
	float totalDepth;
	int managedNum;
	int unmanagedNum;
	float managedDepth;
	float unmanagedDepth;
	int count = 0;

	_getParameters(
		&unmanagedDepth, &managedDepth, &unmanagedNum, &managedNum);

	totalDepth = _rootStart;

	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		bool u = false;
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
    
		if (!win)
		{
			// should not happen
			continue;
		}
		float dd = _getDeltaDepth(
			win, &count, unmanagedDepth, managedDepth, unmanagedNum,
			managedNum);
		totalDepth = totalDepth + dd;
		win->layerCorrection(
			_desktopWidth, _desktopHeight, _eyeDist, totalDepth);
		if (_eyeDist == 0)
		{
			// FIXME: ortho proj
			if (count > 0)
			{
				glTranslatef(0,0,totalDepth);
			}
		}
		(*o)->displayGraph(policy) ;

	}
}

void
LayerManager::select(void)
{
	float totalDepth;
	int managedNum;
	int unmanagedNum;
	float managedDepth;
	float unmanagedDepth;
	int count = 0;

	_getParameters(
		&unmanagedDepth, &managedDepth, &unmanagedNum, &managedNum);

	totalDepth = _rootStart;

	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		bool u = false;
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
    
		if (!win)
		{
			// should not happen
			continue;
		}
		float dd = _getDeltaDepth(
			win, &count, unmanagedDepth, managedDepth, unmanagedNum,
			managedNum);
		totalDepth = totalDepth + dd;
		win->layerCorrection(
			_desktopWidth, _desktopHeight, _eyeDist, totalDepth);
		if (_eyeDist == 0)
		{
			// FIXME: ortho proj
			if (count > 0)
			{
				glTranslatef(0,0,totalDepth);
			}
		}

		(*o)->selectGraph() ;
	}
}

XwncWindow *LayerManager::getUnmanagedWindow(int x, int y)
{
	XwncWindow *rw = 0;

	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		bool ovre = false;
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
    
		if (!win || !win->isUnmanaged())
		{
			continue;
		}
		int wX,wY,wW,wH;
		win->getSize(&wW, &wH);
		win->getRealPosition(&wX, &wY);

		if (x >= wX && x <= wX + wW && y >= wY && y <= wY + wH)
		{
			rw = win;
		}
	}
	return rw;
}

void
LayerManager::setRootWin(sgNode *rootWin)
{
	_rootWin = rootWin;
}

XwncWindow *LayerManager::getRootWindow()
{
	XwncWindow *win = dynamic_cast<XwncWindow *>(_rootWin);
	return win;
}


void LayerManager::restack(sgNode *node, sgNode *nodeNext, bool add)
{
	sgNode *checkNext = 0;
	std::list<sgNode *>::iterator prev;

	if (!add)
	{
		_dependencies.remove(node) ;
	}

	if (nodeNext)
	{
		prev = _dependencies.begin();
		for(std::list<sgNode *>::iterator o=_dependencies.begin();
		    o != _dependencies.end(); 
		    ++o )
		{
			prev = o;
			if ((*o) == nodeNext)
			{
				checkNext = nodeNext;
				break;
			}
		}
	}

	if (checkNext == 0)
	{
		if (nodeNext)
		{
			std::cerr << "restacking: Warnning unknown next window"
				  << std::endl;
		}
		// add at the top of the stacking
		_dependencies.push_back(node) ;
	}
	else
	{
		// add it before checkNext
		_dependencies.insert(prev, node);	
	}

#if 0
	_printStacking();
#endif

	_changed = true ;
}

// for debuging
void
LayerManager::_printStacking(void)
{
	std::cerr << "Stacking: ";

	for(std::list<sgNode *>::iterator o=_dependencies.begin();
	    o != _dependencies.end(); ++o )
	{
		std::cerr << ((XwncWindow *)(*o))->getTitle() << " ";
	}

	std::cerr << std::endl;
}

// --------------------------------------------------------------------
// Inititialisation

void LayerManager::setDesktopProperties(
	float desktopWidth, float desktopHeight, double eyeDist,
	double near, double far, sgViewpoint *viewpoint)
{
	_desktopWidth = desktopWidth;
	_desktopHeight = desktopHeight;
	_eyeDist = eyeDist;
	_rootStart = -(far - eyeDist) + 1;
	_bigGap = (far - near)/5 - 1;
	_viewpoint = viewpoint;
}

LayerManager::LayerManager(std::string name) : sgNode(name)
{
	_desktopWidth = 0.0;
	_desktopHeight = 0.0;
	_eyeDist = 0.0;
	_rootStart = - 0.0;
	_bigGap = 0.0;

	_transX = 0;
	_transY = 0;
	_scaleX = 1;
	_scaleY = 1;

	_surfaceTop = _surfaceLeft = _surfaceBottom = _surfaceRight =
		SURFACE_DEFAULT_CUT;
	_surfaceX = _surfaceY = SURFACE_DEFAULT_CUT;
	_surface = SURFACE_DEFAULT;
	// no surface by default;
	_surfaceOn = false;

	_bg = 0;
	_rootWin = 0;

	_inExposeMode = false;
}

// --------------------------------------------------------------------
// "Operations"

void LayerManager::translate_rel(float x, float y, float z)
{
	float tx = _transX + x;
	float ty = _transY + y;

	float dw = (float)_desktopWidth*_scaleX/2.0;
	float dh = (float)_desktopHeight*_scaleY/2.0;

	float adw = (float)_desktopWidth/2.0;
	float adh = (float)_desktopHeight/2.0;

	float left = -dw + tx*_scaleX;
	float top = dh + ty*_scaleY;
	float right = dw + tx*_scaleX;
	float bottom = -dh + ty*_scaleY;

	if (left > -adw)
	{
		x = (-adw + dw)/_scaleX - _transX;
	}
	if (right < adw)
	{
		x = (adw - dw)/_scaleX - _transX;
	}
	if (top < adh)
	{
		y = (adh - dh)/_scaleY - _transY;
	}
	if (bottom > -adh)
	{
		y = (-adh + dh)/_scaleY - _transY;
	}

	if (x == 0 && y == 0)
	{
		return;
	}
	_transX = _transX + x;
	_transY = _transY + y;
	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
		
		if (!win)
		{
			continue;
		}
		win->translate_rel(x, y, 0);
	}
	XwncWindow::rootTransX = _transX;
	XwncWindow::rootTransY = _transY;
}

void LayerManager::getTranslation(float *x, float *y)
{
	*x = _transX;
	*y = _transY;
}

// X11 geometries!
void LayerManager::x11projectWarp(int x, int y, int *px, int *py, bool warp)
{
	// project with the hands in 3D
	float adw = (float)_desktopWidth/2.0;
	float adh = (float)_desktopHeight/2.0;
	float gx = (((float)x - adw) + _transX)*_scaleX;
	float gy = ((-(float)y + adh) + _transY)*_scaleY;

	if (warp)
	{
		float tx = 0, ty = 0;
		if (gx < -adw)
		{
			tx = (-adw/_scaleX) + adw - _transX - x + 4;
		}
		else if (gx > adw)
		{
			tx = (adw/_scaleX) + adw - _transX - x - 4;
		}
		if (gy > adh)
		{
			ty = (adh/_scaleY) - adh + y - _transY - 4;
		}
		else if (gy < -adh)
		{
			ty = (-adh/_scaleY) - adh + y - _transY + 4;
		}

		if (tx != 0 || ty != 0)
		{
			translate_rel(tx, ty, 0);

			gx = (((float)x - adw) + _transX)*_scaleX;
			gy = ((-(float)y + adh) + _transY)*_scaleY;
		}
	}

	// X11 geometries
	*px = (int)(gx + adw);
	*py = (int)(-gy + adh); 
}

// FIXME: not perfect if the window has some XY rotation
void LayerManager::translateForWindow(XwncWindow *win)
{
	bool found = false;

	if (_scaleX == 1 && _scaleY == 1)
	{
		return;
	}

	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		XwncWindow *w = dynamic_cast<XwncWindow *>(*o);
		if (w == win) found = true; 
	}

	if (!found)
		return;

	float x,y,w,h;
	win->getRealPosition(&x,&y);
	win->getSize(&w,&h);

	// FIXME should not be hard coded
	GLuint selBuffer[3] ;
	GLuint selSize=3;

	selBuffer[2] = (GLuint )(win->getRenderer());
	selBuffer[1] = (GLuint )(win);
	selBuffer[0] = (GLuint )(this);

	float tsX = _scaleX;
	float tsY = _scaleY;

	GLdouble left, top, right, bottom, px, py, pz = 0;
	bool ret;
	ret =_viewpoint->project(
		-w/2, h/2,  0.0, selBuffer, selSize, &left, &top, &pz);
	ret =_viewpoint->project(
		-w/2, -h/2,  0.0, selBuffer, selSize, &px, &bottom, &pz);

	if (px < left)
		left = px;

	ret =_viewpoint->project(
		w/2, -h/2,  0.0, selBuffer, selSize, &right, &py, &pz);
	if (py < bottom)
		bottom = py;
	ret =_viewpoint->project(
		w/2, h/2,  0.0, selBuffer, selSize, &px, &py, &pz);
	if (py > top)
		top = py;
	if (px > right)
		right = px;

	// upside down? ... FIXME?
	if (top < bottom)
	{
		py = top;
		top = bottom;
		bottom = py;
	}
	if (right < left)
	{
		px = right;
		right = left;
		left = px;
	}

#if 0
	std::cerr << "Translate: " << left << " " << top << " " << right << " "
		  << bottom <<  std::endl;
#endif

	float tx = 0, ty = 0;
	if (left < 0)
	{
		tx = - left / tsX;
	}
	else if (right > _desktopWidth)
	{
		tx = (_desktopWidth - right)/tsX;
		left = left + tx*tsX;
		if (left < 0)
		{
			tx += - left / tsX;
		}
	}
	if (top > _desktopHeight)
	{
		ty = (_desktopHeight - top) / tsY;
	}
	else if (bottom < 0)
	{
		ty = - bottom / tsY;
		top = top + ty*tsY;
		if (top > _desktopHeight)
		{
			ty += (_desktopHeight - top)/ tsY;
		}
	}

	translate_rel(tx, ty, 0);
}

// simplistic
unsigned int LayerManager::_guessGravity(XwncWindow *win)
{
	unsigned int gravity = NorthWestGravity; // default
	unsigned int NSGravity = NorthGravity; // default
	unsigned int WEGravity = WestGravity; // default

	float x,y,width,height;
	win->getRealPosition(&x,&y);
	win->getSize(&width,&height);
	
	float factorW = _desktopWidth/10;
	float leftLimit = _desktopWidth/4;
	float leftStrongLimit = factorW;
	float rightLimit = _desktopWidth - leftLimit;
	float rightStrongLimit = _desktopWidth - leftStrongLimit;

	float factorH = _desktopHeight/10;
	float topLimit = _desktopHeight/4;
	float topStrongLimit = factorH;
	float bottomLimit = _desktopHeight - topLimit;
	float bottomStrongLimit = _desktopHeight - topStrongLimit;

	if (x >= rightLimit)
	{
		WEGravity = EastGravity;
	}
	else if (x >= leftLimit)
	{
		if (x + width >= rightStrongLimit)
			WEGravity = EastGravity;
		else
			WEGravity = CenterGravity;
	}
	else if (x > leftStrongLimit)
	{
		if (x + width > rightStrongLimit)
			WEGravity = EastGravity;
		else if (x + width > leftLimit)
			WEGravity = CenterGravity;
	}
	else if (x + width > rightStrongLimit) // && x <= leftStrongLimit
	{
		if (_desktopWidth - (x + width) < x)
		    WEGravity = EastGravity;
		else if (_desktopWidth - (x + width) == x)
		    WEGravity = CenterGravity;
	}

	if (y >= bottomLimit)
	{
		NSGravity = SouthGravity;
	}
	else if (y >= topLimit)
	{
		if (y + height >= bottomStrongLimit)
			NSGravity = SouthGravity;
		else
			NSGravity = CenterGravity;
	}
	else if (y > topStrongLimit)
	{
		if (y + height > bottomStrongLimit)
			NSGravity = SouthGravity;
		else if (y + height > bottomLimit)
			NSGravity = CenterGravity;
	}
	else if (y + height >= bottomStrongLimit) // && y <= topStrongLimit
	{
		if (_desktopHeight - (y + height) < y)
		    NSGravity = SouthGravity;
		else if (_desktopHeight - (x + width) == x)
		    NSGravity = CenterGravity;
	}

	switch(NSGravity)
	{
	case NorthGravity:
	{
		switch(WEGravity)
		{
		case EastGravity:
			gravity = NorthEastGravity;
			break;
		case CenterGravity:
			gravity = NorthGravity;
			break;
		case WestGravity:
			gravity = NorthWestGravity;
			break;
		default:
			break;
		}
	}
	break;
	case CenterGravity:
	{
		switch(WEGravity)
		{
		case EastGravity:
			gravity = EastGravity;
			break;
		case CenterGravity:
			gravity = CenterGravity;
			break;
		case WestGravity:
			gravity = WestGravity;
			break;
		default:
			break;
		}
	}
	break;
	case SouthGravity:
	{
		switch(WEGravity)
		{
		case EastGravity:
			gravity = SouthEastGravity;
			break;
		case CenterGravity:
			gravity = SouthGravity;
			break;
		case WestGravity:
			gravity = SouthWestGravity;
			break;
		default:
			break;
		}
	}
	break;
	default:
		break;
	}

	return gravity;
}

void LayerManager::scaleWindow(
	XwncWindow *win, GLfloat fx, GLfloat fy, GLfloat fz,
	unsigned int gravity, bool absolute)
{
	float x,y,w,h;
	win->getRealPosition(&x,&y);
	win->getSize(&w,&h);

	// FIXME should not be hard coded
	GLuint selBuffer[3] ;
	GLuint selSize=3;

	selBuffer[2] = (GLuint )(win->getRenderer());
	selBuffer[1] = (GLuint )(win);
	selBuffer[0] = (GLuint )(this);

	float tsX = _scaleX;
	float tsY = _scaleY;

	if (gravity == 0)
	{
		gravity = _guessGravity(win);
	}

	GLdouble left, top, gx, gy, qx, qy, rx, ry, px, py, pz = 0;
	bool ret;
	switch (gravity)
	{
	case SouthEastGravity:
		gx = w/2; gy = -h/2;
		break;
	case SouthGravity:
		gx = 0; gy = -h/2;
		break;
	case SouthWestGravity:
		gx = -w/2; gy = -h/2;
		break;
	case EastGravity:
		gx = w/2; gy = 0;
		break;
	case CenterGravity:
		gx = 0; gy = 0;
		break;
	case WestGravity:
		gx = -w/2; gy = 0;
		break;
	case NorthEastGravity:
		gx = w/2; gy = h/2;
		break;
	case NorthGravity:
		gx = 0; gy = h/2;
		break;
	case NorthWestGravity:
	default:
		gx = -w/2; gy = h/2;
		break;
	}

	ret =_viewpoint->project(
		-w/2, h/2,  0.0, selBuffer, selSize, &left, &top, &pz);

	if (gravity == NorthWestGravity)
	{
		qx = left;
		qy = top;
	}
	else
	{
		ret =_viewpoint->project(
			gx, gy,  0.0, selBuffer, selSize, &qx, &qy, &pz);
	}

	if (absolute)
	{
		win->absoluteScale(fx, fy, fz);
	}
	else
	{
		win->getRenderer()->scale(fx, fy, fz);
	}

	ret =_viewpoint->project(
		-w/2, h/2,  0.0, selBuffer, selSize, &px, &py, &pz);
	if (gravity == NorthWestGravity)
	{
		rx = px;
		ry = py;
	}
	else
	{
		ret =_viewpoint->project(
			gx, gy,  0.0, selBuffer, selSize, &rx, &ry, &pz);
	}
	win->internalTranslation((-rx + qx)/_scaleX, (- ry + qy)/_scaleY, 0);

	// FIXME: keep on screen (top,left)?
}

void LayerManager::rotateWindow(
	XwncWindow *win, GLfloat angle, GLfloat fx, GLfloat fy, GLfloat fz,
	unsigned int gravity, bool absolute)
{
	float x,y,w,h;
	win->getRealPosition(&x,&y);
	win->getSize(&w,&h);

	// FIXME should not be hard coded
	GLuint selBuffer[3] ;
	GLuint selSize=3;

	selBuffer[2] = (GLuint )(win->getRenderer());
	selBuffer[1] = (GLuint )(win);
	selBuffer[0] = (GLuint )(this);

	float tsX = _scaleX;
	float tsY = _scaleY;

	if (gravity == 0)
	{
		gravity = _guessGravity(win);
	}

	GLdouble left, top, gx, gy, qx, qy, rx, ry, px, py, pz = 0;
	bool ret;
	switch (gravity)
	{
	case SouthEastGravity:
		gx = w/2; gy = -h/2;
		break;
	case SouthGravity:
		gx = 0; gy = -h/2;
		break;
	case SouthWestGravity:
		gx = -w/2; gy = -h/2;
		break;
	case EastGravity:
		gx = w/2; gy = 0;
		break;
	case CenterGravity:
		gx = 0; gy = 0;
		break;
	case WestGravity:
		gx = -w/2; gy = 0;
		break;
	case NorthEastGravity:
		gx = w/2; gy = h/2;
		break;
	case NorthGravity:
		gx = 0; gy = h/2;
		break;
	case NorthWestGravity:
	default:
		gx = -w/2; gy = h/2;
		break;
	}

	ret =_viewpoint->project(
		-w/2, h/2,  0.0, selBuffer, selSize, &left, &top, &pz);

	if (gravity == NorthWestGravity)
	{
		qx = left;
		qy = top;
	}
	else
	{
		ret =_viewpoint->project(
			gx, gy,  0.0, selBuffer, selSize, &qx, &qy, &pz);
	}

	if (absolute)
	{
		win->absoluteRotate(angle, fx, fy, fz);
	}
	else
	{
		win->getRenderer()->rotate_rel(angle, fx, fy, fz);
	}

	ret =_viewpoint->project(
		-w/2, h/2,  0.0, selBuffer, selSize, &px, &py, &pz);
	if (gravity == NorthWestGravity)
	{
		rx = px;
		ry = py;
	}
	else
	{
		ret =_viewpoint->project(
			gx, gy,  0.0, selBuffer, selSize, &rx, &ry, &pz);
	}
	win->internalTranslation((-rx + qx)/_scaleX, (- ry + qy)/_scaleY, 0);

	// FIXME: keep on screen (top,left)?
}

void LayerManager::scale(float x, float y, float z)
{
	float sx = _scaleX*x;
	float sy = _scaleX*y;

#if 1
	if (sx < 1)
	{
		if (_scaleX > 1)
		{
			x = 1/_scaleX;
		}
		else
		{
			x = 1;
		}
	}
	if (sy < 1)
	{
		if (1||_scaleY > 1)
		{
			y = 1/_scaleY;
		}
		else
		{
			y = 1;
		}
	}
#endif
	_scaleX = _scaleX*x;
	_scaleY = _scaleY*y;
	sgNode::scale(x, y, 1);
	translate_rel(0,0,0);
}

void LayerManager::resetTransformations(void)
{
	sgNode::resetTransformations();
	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
		
		if (!win)
		{
			continue;
		}
		win->translate_rel(-_transX, -_transY, 0);
	}
	_transX = 0;
	_transY = 0;
	_scaleX = 1;
	_scaleY = 1;
	XwncWindow::rootTransY = 0;
	XwncWindow::rootTransX = 0;
}

// --------------------------------------------------------------------
// for click motion

void LayerManager::setTmpAlphaAbove(
	XwncWindow *startWindow, bool set, float alpha)
{
	bool start = false; 

	for( std::list<sgNode *>::iterator o=_dependencies.begin();
	     o != _dependencies.end(); 
	     ++o )
	{
		XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
    
		if (!win || win->isUnmanaged())
		{
			continue;
		}
		if (win == startWindow)
		{
			start = true;
			continue;
		}
		if (!start)
		{
			continue;
		}
		if (!set)
		{
			win->resetAlpha();
		}
		else
		{
			win->setTmpAlpha(alpha);
		}
		
	}
}

// --------------------------------------------------------------------
// Surface

void LayerManager::setSurfaceParameters(
	float surfaceLeft, float surfaceTop,
	float surfaceRight, float surfaceBottom)
{
	_surfaceTop = surfaceTop;
	_surfaceLeft = surfaceLeft;
	_surfaceBottom = surfaceBottom;
	_surfaceRight = surfaceRight;

	_changed = true;
}

void LayerManager::setSurfaceParameters(float cutX, float cutY)
{
	setSurfaceParameters(cutX, cutY, cutX, cutY);
}

void LayerManager::setSurfaceParameters(float cut)
{
	setSurfaceParameters(cut, cut, cut, cut);
}

// surface by name
void LayerManager::setSurface(int surface, float cutX, float cutY)
{
	_surfaceX = cutX;
	_surfaceY = cutY;
	_surface = surface;
	switch(surface)
	{
	case SURFACE_TBCYLINDER:
		_surfaceTop = cutY;
		_surfaceLeft = 0;
		_surfaceBottom = cutY; 
		_surfaceRight = 0;
		break;
	case SURFACE_LRCYLINDER:
		_surfaceTop = 0;
		_surfaceLeft = cutX;
		_surfaceBottom = 0; 
		_surfaceRight = cutX;
		break;
	case SURFACE_SPHERE_CYLINDER:
		_surfaceTop = cutY;
		_surfaceLeft = cutX;
		_surfaceBottom = 0; 
		_surfaceRight = cutX;
		break;
	case SURFACE_CYLINDER_SPHERE:
		_surfaceTop = 0;
		_surfaceLeft = cutX;
		_surfaceBottom = cutY; 
		_surfaceRight = cutX;
		break;
	default:
	case SURFACE_SPHERE:
		_surfaceTop = cutY;
		_surfaceLeft = cutX;
		_surfaceBottom = cutY; 
		_surfaceRight = cutX;
		break;
	}

	_changed = true;
}

void LayerManager::setSurface(int surface, float cut)
{
	setSurface(surface, cut, cut);
}

void LayerManager::setSurface(int surface)
{
	setSurface(surface, _surfaceX, _surfaceY);
}

void LayerManager::setSurface(float cutX, float cutY)
{
	setSurface(_surface, cutX, cutY);
}

void LayerManager::setSurface(float cut)
{
	setSurface(_surface, cut, cut);
}

void LayerManager::enableSurface(int toggle)
{
	switch(toggle)
	{
	case 0:
		_surfaceOn = false;
		break;
	case 1:
		_surfaceOn = true;
		break;
	case -1:
	default:
		_surfaceOn = !_surfaceOn;
		break;
	}
	_changed = true;
}

void LayerManager::getSurfaceRotation(float x, float y, double *rx, double *ry)
{
	if (!_surfaceOn)
	{
		*rx = 0.0;
		*ry = 0.0;
		return;
	}

	double dw = _desktopWidth/2;
	double dh = _desktopHeight/2;

	if (x > dw)
	{
		*rx = _surfaceRight * 90;
	}
	else if (x < - dh)
	{
		*rx = - _surfaceLeft * 90;
	}
	else if (x >= 0)
	{
		*rx = _surfaceRight * (x/dw) * 90;	
	}
	else
	{
		*rx = _surfaceLeft * (x/dh) * 90;
	}

	if (y > dh)
	{
		*ry = _surfaceTop * 90;
	}
	else if (y < - dh)
	{
		*ry = - _surfaceBottom * 90;
	}
	else if (y >= 0)
	{
		*ry = _surfaceTop  * (y/dh) * 90;	
	}
	else
	{
		*ry = _surfaceBottom * (y/dh) * 90;
	}
}

// --------------------------------------------------------------------
// Expose

bool LayerManager::inExposeMode(void)
{
	return _inExposeMode;
}

void LayerManager::exposeMode(void)
{
	static float stransX = 0, stransY = 0, sscaleX = 1, sscaleY = 1;

	if (!_inExposeMode)
	{
		int count = 0;
		_inExposeMode = true;
		float adw = (float)_desktopWidth/2.0;
		float adh = (float)_desktopHeight/2.0;

		sgNode::saveTransformations();
		stransX = _transX;
		stransY = _transY;
		sscaleX = _scaleX;
		sscaleY = _scaleY;
		for( std::list<sgNode *>::iterator o=_dependencies.begin();
		     o != _dependencies.end(); 
		     ++o )
		{
			XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
			win->getRenderer()->saveTransformations();
		}
		resetTransformations();
		for( std::list<sgNode *>::iterator o=_dependencies.begin();
		     o != _dependencies.end(); 
		     ++o )
		{
			XwncWindow *win = dynamic_cast<XwncWindow *>(*o);

			win->getRenderer()->resetTransformations();
			if (!win->isRootWindow() && !win->isEwmhDesktop() &&
			    !win->isUnmanaged())
			{
				count++;
			}
		}

		// FIXME: implement a better algo

		float row = 1, column = 1; 
		while (row*column < count)
		{
			if (column <= row) column++;
			else if (row <= column) row++;
		}
		float height = (float)_desktopHeight/row;
		float width = (float)_desktopWidth/column;
		int r = 1, c = 1;
		float xx = 0;
		float yy = 0;
		for( std::list<sgNode *>::iterator o=_dependencies.begin();
		     o != _dependencies.end(); 
		     ++o )
		{
			XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
	
			if (win->isRootWindow() || win->isEwmhDesktop() ||
			    win->isUnmanaged())
			{
				// in the void
				win->getRenderer()->translate_rel(
					0, 0, 2*_rootStart);
				continue;
			}

			float wH,wW;
			win->getSize(&wW,&wH);

			float sw = 1, sh = 1;
			if (wW > width) sw = width/wW;
			if (wH > height) sh = height/wH;
			if (sh < sw) sw = sh;

			float tx,ty;

			tx = (xx + width/2) - (_desktopWidth/2);
			ty = -(yy + height/2) + (_desktopHeight/2);
			win->getRenderer()->translate_rel(2*tx, 2*ty, -_eyeDist);
			win->getRenderer()->scale(2*sw, 2*sw, 1);

			if (c < column)
			{
				c++;
				xx = xx + width;
			}
			else
			{
				c = 1;
				xx = 0;
				yy = yy + height;
				r++;
			}
		}
	}
	else
	{
		_inExposeMode = false;
		// restore;
		for( std::list<sgNode *>::iterator o=_dependencies.begin();
		     o != _dependencies.end(); 
		     ++o )
		{
			XwncWindow *win = dynamic_cast<XwncWindow *>(*o);
			win->getRenderer()->restoreSavedTransformations();
		}
		sgNode::restoreSavedTransformations();
		_scaleX = sscaleX;
		_scaleY = sscaleY;
		_transX = stransX;
		_transY = stransY;
		XwncWindow::rootTransY = _transY;
		XwncWindow::rootTransX = _transX;
	}
}
