/*
 *
 * main/ametista.cxx --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include <sys/time.h>

#include <nucleo/core/ReactiveEngine.H>

#include <nucleo/utils/AppUtils.H>
#include <nucleo/utils/FileUtils.H>

#include <nucleo/gl/window/glWindow.H>
#include <nucleo/image/sink/ImageSink.H>

#include "wstyle/AbstractWindowRenderer.H"
#include "wncdesktop/XwncWindow.H"
#include "wncdesktop/XwncDesktop.H"
#include "fvwmmodule/FvwmModule.H"

#include "AUtils.H"
#include "AScreen.H"

#include <sstream>

using namespace nucleo ;

#define VERBOSE 0

void
quitFvwmAndExit(int fds[2]) {
  SendText(fds, "Quit", 0) ;
  FvwmPacket* packet = ReadFvwmPacket(fds[1]) ;
  if (packet!=NULL) std::cerr << "quitFvwmAndExit: received a non-NULL packet" << std::endl ;
  exit(0) ;
}

int
main(int argc, char **argv)
{
#if VERBOSE
  std::cerr << "argc=" << argc << " argv=[" << argv[0] ;
  for (int ia=1; ia<argc; ia++) 
    std::cerr << " " << argv[ia] ;
  std::cerr << "]" << std::endl ;
#endif

  /*
    From http://www.fvwm.org/documentation/dev_modules.php
    
    argv[0] is the full pathname of the module
    argv[1] and argv[2] are file descriptors that can be used to communicate with fvwm
    argv[3] is the full path name of the configuration file from which the module was started
    argv[4] is the application window in whose context the module was launched
    argv[5] is the context of the window decoration in which the module was launched

    argv[6] and above are user specified command line arguments
  */

  char *myName = GetFileNameFromPath(argv[0]);

  if(argc<6) {
    std::cerr << myName << ": this program can be run only as an FVWM module" << std::endl;
    exit(1) ;
  }

  int fvwm_fds[2] ;
  fvwm_fds[0] = atoi(argv[1]) ;
  fvwm_fds[1] = atoi(argv[2]) ;

  std::stringstream rsBuffer ;
  rsBuffer << "AFuncRestartModule " << myName ;
  for (int ia=6; ia<argc; ++ia) rsBuffer << " " << argv[ia] ;
  std::string restartString=rsBuffer.str() ;
#if VERBOSE
  std::cerr << "restartString = [" << restartString << "]" << std::endl ;
#endif

  int user_argc = argc-5 ;
  char **user_argv = new char * [user_argc] ;
  user_argv[0] = strdup(argv[0]) ;
  for (int uia=1; uia<user_argc; ++uia) user_argv[uia] = strdup(argv[5+uia]) ;
#if VERBOSE
  std::cerr << "uargc=" << user_argc << " uargv=[" << user_argv[0] ;
  for (int uia=1; uia<user_argc; uia++) 
    std::cerr << " " << user_argv[uia] ;
  std::cerr << "]" << std::endl ;
#endif

  bool SHOW_CURSOR = false ;
  bool SHOW_FPS = false ;
  bool FOCUS_HILITE = false ;
  bool SHADOW = false ;
  char *CAPTURE = 0 ;
  bool ORTHO = false ;
  bool SHAPED = false ;
  bool NOSTENCIL = false;
  char *DISPLAY = 0 ;
  char *WNC_URI = 0 ;
  bool USE_MIPMAPS = false;
  bool USE_RGB_TEX = false;

  // NPOT_EXT is disabled by default as XFree-4.4.0 (rc1) has this
  // extension but it is totally broken with the radeon driver
  bool USE_NPOT_EXT = false ;
  bool NO_SHM = false;

  if (parseCommandLine(
	      user_argc, user_argv, "CDHsc:oSd:w:nmtpa", "bbbbsbbssbbbbb",
	      &SHOW_CURSOR, &SHOW_FPS, &FOCUS_HILITE, &SHADOW, &CAPTURE, &ORTHO,
	      &SHAPED, &DISPLAY, &WNC_URI, &USE_NPOT_EXT,&NO_SHM,&NOSTENCIL,
	      &USE_MIPMAPS,&USE_RGB_TEX) < 0)
  {	
	  std::cerr << std::endl << myName ;
	  std::cerr << " [-C (software cursor)] [-D(isplay fps)] "
		    << "[-H (hilite focused window)] "
		    << "[-s (enable window shadow)] " 
		    << "[-c capture-sink] [-o(rthogonal projection)] "
		    << "[-d display] [-w WNC-URI] "
		    << "[-n (use npot texture extension)] "
		    << "[-m (do not use shared memory)] "
		    << "[-t (do not use the stencil buffer)] "
		    << "[-a (use RGB texture in place of RGBA texture)] "
		    << std::endl ;
	  std::cerr << "WNC URI and display MUST be specified" << std::endl ;
	  std::cerr << std::endl << std::endl ;
	  quitFvwmAndExit(fvwm_fds) ;
  }

  if (!WNC_URI || !DISPLAY)
  {
    std::cerr << "w=" << WNC_URI << " d=" << DISPLAY << std::endl ;
    std::cerr << std::endl << myName << ": " ;
    std::cerr << "WNC URI and display MUST be specified (using -w and -d)" ;
    std::cerr << std::endl << std::endl ;
    quitFvwmAndExit(fvwm_fds) ;
  }

  // ------------------------------------------------------
  // Connect to the Xwnc server via a XwncDesktop

  XwncDesktop wncDesktop(
	  WNC_URI, "foldable", (NO_SHM)? false:true, USE_RGB_TEX) ;

  // ------------------------------------------------------

  char *env;
  env = (char *)malloc(9 + strlen(DISPLAY));
  env[0] = 0;
  strcat(env, "DISPLAY=");
  strcat(env, DISPLAY);
  putenv(env);
  
  // FIXME: disable ortho projection as it is totally broken;
  // should be fixed.
  if (ORTHO) {
	  std::cerr << std::endl << myName << ": orthogonal projection is temporally disabled" << std::endl;
	  ORTHO = false;
  }

  if (!USE_NPOT_EXT)
  {
	  glTiledTexturedImage::use_npot_ext = false;
  }
  if (USE_MIPMAPS)
  {
	  glTiledTexturedImage::do_generate_mipmaps = true;
  }

  // ------------------------------------------------------

  AScreen *AScr = new AScreen(
	  &wncDesktop, ORTHO, SHAPED, FOCUS_HILITE, NOSTENCIL) ;
  AScr->setTitle(myName) ;

  wncDesktop.setAScreen(AScr);

  {
    unsigned int w,h;
    wncDesktop.getSize(&w, &h);
    AScr->setGeometry(w, h);
  }

  AScr->makeCurrent() ;

  if (CAPTURE) AScr->initCapture(CAPTURE) ;

  // ------------------------------------------------------

  FvwmModule fvwmModule(AScr, &wncDesktop, fvwm_fds[0], fvwm_fds[1], myName, restartString) ;

  glClearColor(0.1,0.5,0.4,1) ;

  if (SHOW_FPS && !SHAPED) AScr->showFPS();

  if (!FOCUS_HILITE) AbstractWindowRenderer::dimmedByDefault = false ;

  if (!SHADOW) AbstractWindowRenderer::shadowByDefault = false ;

  // ------------------------------------------------------

  srandom((unsigned int)(time(0)%737)) ;

  AScr->setTransparent(true);

#ifdef __APPLE__
  SHOW_CURSOR = true;
#endif
  if (SHOW_CURSOR) AScr->showCursor();
  // ------------------------------------------------------

  for (;;) {
    ReactiveEngine::step() ;
    if (!AScr->handleEvents()) break ;
    AScr->draw(false) ;
  }

  if (CAPTURE) AScr->deleteCapture() ;

  return 0 ;
}
