/*
 *
 * wnc/wncSource.cxx --
 *
 * Copyright (C) Nicolas Roussel, Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include <nucleo/core/ReactiveEngine.H>
//#include <nucleo/image/processing/basic/PaintingAndTransforming.H>
#include <nucleo/image/encoding/Conversion.H>
#include <nucleo/gl/window/keysym.H>
#include <nucleo/utils/FileUtils.H>

#include <X11/Xmd.h>
#include <X11/Xutil.h>

extern "C" {
#include <wncproto.h>
#include <wncauth.h>
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/ipc.h> 
#include <sys/shm.h>
}

#include <unistd.h>
#include <stdio.h>
#include <math.h>

#include <stdexcept>

#include "XwncDesktop.H"
#include "XwncWindow.H"
#include "wncSource.H"


/*
 * NOTE: VNC supports bitsPerPixel of 8, 16 and 32 only (not 24).
 * at 32 (24) RGB is in fact raw RGBA and at 16 it is RGB565
 *
 * The alpha is broken in any case. 
 */

/*
 * When leaving a window, should we generate keyRelease events for
 * pressed modifiers (VNC client does so) ?
 *
 * Yes we do that
 *
 */

#define DEBUG_LEVEL 1

#ifdef WORDS_BIGENDIAN
#define Swap16IfLE(s) (s)
#else 
#define Swap16IfLE(s) ((((s) & 0xff) << 8) | (((s) >> 8) & 0xff)) 
#endif

#ifdef WORDS_BIGENDIAN
#define Swap32IfLE(l) (l)
#else 
#define Swap32IfLE(l) \
     ((((l) & 0xff000000) >> 24)  | \
      (((l) & 0x00ff0000) >>  8)  | \
      (((l) & 0x0000ff00) <<  8)  | \
      (((l) & 0x000000ff) << 24))
#endif

#if DEBUG_LEVEL>=1
static void
PrintPixelFormat(rfbPixelFormat *format) {
  if (format->bitsPerPixel == 1) {
    fprintf(stderr,"  Single bit per pixel.\n");
    fprintf(stderr,
		  "  %s significant bit in each byte is leftmost on the screen.\n",
		  (format->bigEndian ? "Most" : "Least"));
  } else {
    fprintf(stderr,"  %d bits per pixel.\n",format->bitsPerPixel);
    if (format->bitsPerPixel != 8) {
      fprintf(stderr,"  %s significant byte first in each pixel.\n",
		    (format->bigEndian ? "Most" : "Least"));
    }
    if (format->trueColour) {
      fprintf(stderr,"  True colour: max red %d green %d blue %d",
		    format->redMax, format->greenMax, format->blueMax);
      fprintf(stderr,", shift red %d green %d blue %d\n",
		    format->redShift, format->greenShift, format->blueShift);
    } else {
      fprintf(stderr,"  Colour map (not true colour).\n");
    }
  }
}
#endif


#define WNC_TIME_DEBUG 0

#define WNC_TIME_DEBUG_FRAME_ONLY 0

#define WNC_TIME_DEBUG_START      0
#define WNC_TIME_DEBUG_FBU_START  1
#define WNC_TIME_DEBUG_FBU_STOP   2
#define WNC_TIME_DEBUG_FRAME_START  3
#define WNC_TIME_DEBUG_FRAME_STOP   4
#define WNC_TIME_DEBUG_BITS   5

#if WNC_TIME_DEBUG
#include <sys/time.h>

static 
int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y)
{
       /* Perform the carry for the later subtraction by updating Y. */
       if (x->tv_usec < y->tv_usec) {
         int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
         y->tv_usec -= 1000000 * nsec;
         y->tv_sec += nsec;
       }
       if (x->tv_usec - y->tv_usec > 1000000) {
         int nsec = (x->tv_usec - y->tv_usec) / 1000000;
         y->tv_usec += 1000000 * nsec;
         y->tv_sec -= nsec;
       }
     
       /* Compute the time remaining to wait.
          `tv_usec' is certainly positive. */
       result->tv_sec = x->tv_sec - y->tv_sec;
       result->tv_usec = x->tv_usec - y->tv_usec;
     
       /* Return 1 if result is negative. */
       return x->tv_sec < y->tv_sec;
}
#endif


static void
bitsPerSeconds(int bits, int type, bool shm)
{
#if WNC_TIME_DEBUG
	static unsigned long totalBits = 0;
	static unsigned long shmBits = 0;
	static unsigned long rectangles = 0;
	static unsigned long updates = 0;
	static unsigned long frames = 0;
	static struct timeval time;
	struct timeval currentTime;

	if (type == WNC_TIME_DEBUG_START)
	{
		gettimeofday(&time, NULL);
		return;
	}
	else if (type == WNC_TIME_DEBUG_FBU_STOP)
	{
		updates++;
		return;
	}
	else if (type == WNC_TIME_DEBUG_BITS)
	{
		rectangles++;
		totalBits += bits;
		if (shm)
		{
			shmBits += bits;
		}
	}

	if (type != WNC_TIME_DEBUG_FRAME_STOP)
	{
		return;
	}

	frames++;
	gettimeofday(&currentTime, NULL);

	if (currentTime.tv_sec - time.tv_sec >= 5)
	{
		timeval diff;
		float tmicro, tsec, tmilli;
		
		timeval_subtract(&diff, &currentTime, &time);

		tmicro = diff.tv_sec * 1000000 + diff.tv_usec;
		tmilli = tmicro / 1000.0;
		tsec = tmicro / 1000000.0;

#if WNC_TIME_DEBUG_FRAME_ONLY
		std::cerr
			<< "--- " << "time: " << tsec << " seconds  "
			<< " real updates/sec: " << (float)(frames)/tsec
			<< std::endl;
#else
		std::cerr
			<< "--- " << "time: " << tsec << " seconds" << std::endl
			<< "bits: " << totalBits
			<< " (shm: " << ((shmBits)?1:0)
			<< ") rectangles: " << rectangles
			<< " updates: " << updates 
			<< " real updates: " << frames 
			  << std::endl
			  << "bits/millisecond: "
			  << (float)totalBits/tmilli
			  << " real updates/sec: "
			  << (float)(frames)/tsec
			  << " rectangles/frame: "
			  << (float)rectangles/(float)frames
			  << std::endl;
#endif		
		gettimeofday(&time, NULL);
		frames = totalBits = rectangles = shmBits = updates = 0;
		
	}
#endif

}

// ----------------------------------------------------------------------
#define BUF_SIZE 8192
static char buf[BUF_SIZE];
static char *bufoutptr = buf;
static int buffered = 0;

bool
wncSource::_ReadFromRFBServer(char *out, unsigned int n)
{
	int rfbsock = _conn->getFd() ;

	if (n <= buffered)
	{
		memcpy(out, bufoutptr, n);
		bufoutptr += n;
		buffered -= n;
		return True;
	}

	memcpy(out, bufoutptr, buffered);

	out += buffered;
	n -= buffered;

	bufoutptr = buf;
	buffered = 0;

	if (n <= BUF_SIZE)
	{

		while (buffered < n)
		{
			int i = read(
				rfbsock, buf + buffered, BUF_SIZE - buffered);
			if (i <= 0)
			{
				if (i < 0)
				{
					if (errno == EWOULDBLOCK ||
					    errno == EAGAIN) {
						std::cerr <<
							"Should process event?"
							  << std::endl;
						i = 0;
					}
					else
					{
						perror(": read");
						return False;
					}
				}
				else
				{
					if (1)
					{
						std::cerr <<
							"WNC server closed " <<
							"connection"
							  << std::endl;
					}
					return False;
				}
			}
			buffered += i;
		}

		memcpy(out, bufoutptr, n);
		bufoutptr += n;
		buffered -= n;
		return True;
	}
	else
	{
		
		while (n > 0)
		{
			int i = read(rfbsock, out, n);
			if (i <= 0)
			{
				if (i < 0)
				{
					if (errno == EWOULDBLOCK ||
					    errno == EAGAIN) {
						std::cerr <<
							"Should process event?"
							  << std::endl;
						i = 0;
					}
					else
					{
						perror(": read");
						return False;
					}
				}
				else
				{
					if (1)
					{
						std::cerr << "WNC server closed "
							  << "connection"
							  << std::endl;
					}
					return False;
				}
			}
			out += i;
			n -= i;
		}

		return True;
	}
}

void
wncSource::_receive(char *data, unsigned int length) {
	int bytestoread = length-_buffer.length() ;
	if (bytestoread>0)
	{
		int fd = _conn->getFd() ;
		int buffersize = bytestoread ;
		if (buffersize<64000) buffersize = 64000 ;
		char *tmpbuffer = new char [buffersize] ;
		int bytesread = 0 ;
		while (bytesread<bytestoread)
		{
			int newbytes = read(
				fd, (char *)(tmpbuffer+bytesread),
				buffersize-bytesread) ;
			if (newbytes==-1)
			{
				throw std::runtime_error(
					"wncSource: Connection closed");
			}
			bytesread += newbytes ;

#if 0
			std::cerr << newbytes << " new bytes (" << bytestoread
				  << ", " << bytestoread-bytesread << " to go)"
				  << std::endl ;
#endif

		}
		_buffer.append((char *)tmpbuffer, bytesread) ;
		delete [] tmpbuffer ;
	}

#if 0
	std::cerr << "Reading " << length << " bytes from buffer" << std::endl ;
#endif

	memcpy(data, _buffer.c_str(), length) ;
	_buffer.erase(0, length) ;
}

//#define RECEIVE(a,b) _ReadFromRFBServer(a,b);
#define RECEIVE(a,b) _receive(a,b);

// ------------------------------------------------------------------

wncSource::wncSource(
	XwncDesktop *wncDesktop,  bool force_rgb_tex,
	const URI& uri, bool use_shm)
{
	_wncDesktop = wncDesktop;
	_hostname = uri.getHost() ;
	_port = 5900+atoi(uri.getPort().c_str()) ;
	std::string query = uri.getQuery() ;
	_password = URI::getQueryArg(query, "password") ;
	_conn = 0 ;
	_useShm = use_shm;
	// computed later
	_texEncoding = Image::OPAQUE;
	_wncEncoding = Image::OPAQUE;
	_depth = 0;
	if (force_rgb_tex)
	{
		_texEncoding = Image::RGB;
	}
	// testing
	//_texEncoding = Image::RGBA;
	//_texEncoding = Image::ARGB;
	//_texEncoding = Image::RGB;
	//_texEncoding = Image::RGB565;
}

bool
wncSource::start(void)
{
	try {
		_conn = new TcpConnection(_hostname.c_str(), _port) ;

		// --------------------------------------------------------
		// Greetings

		rfbProtocolVersionMsg pv;
		RECEIVE(pv, sz_rfbProtocolVersionMsg) ;
		pv[sz_rfbProtocolVersionMsg] = 0 ;

		int major,minor;
		if (sscanf(pv,rfbProtocolVersionFormat,&major,&minor) != 2)
		{
			throw std::runtime_error(
				"wncSource: not a valid VNC server") ;
		}

#if DEBUG_LEVEL>=1
		std::cerr << "wncSource: WNC server supports protocol version "
			  << major << "." << minor ;
		std::cerr << "(viewer " << rfbProtocolMajorVersion << "."
			  << rfbProtocolMinorVersion << ")" << std::endl ;
#endif

		major = rfbProtocolMajorVersion ;
		minor = rfbProtocolMinorVersion ;
		sprintf(pv,rfbProtocolVersionFormat,major,minor) ;
		_conn->send(pv, sz_rfbProtocolVersionMsg) ;

		// -----------------------------------------------------------
		// Authentication

		CARD32 authScheme ;
		RECEIVE((char *)&authScheme, 4) ;
		authScheme = Swap32IfLE(authScheme);

		switch (authScheme)
		{
		case rfbConnFailed:
		{
			CARD32 reasonLen ;
			RECEIVE((char *)&reasonLen, 4) ;
			reasonLen = Swap32IfLE(reasonLen) ;
			char *reason = new char [reasonLen] ;
			RECEIVE(reason, reasonLen) ;
			std::string msg = "wncSource: WNC connection failed (" ;
			msg.append(reason, reasonLen) ;
			msg.append(")") ;
			throw std::runtime_error(msg) ;
		}
		break ;

		case rfbNoAuth:
#if DEBUG_LEVEL>=1
			std::cerr << "wncSource: No authentication needed"
				  << std::endl ;
#endif
			break;

		case rfbVncAuth:
		{
			CARD8 challenge[CHALLENGESIZE];
			RECEIVE((char *)challenge, CHALLENGESIZE) ;

			if (_password!="")
			{
				vncEncryptBytes(
					challenge, (char *)_password.c_str()) ;
			}
			else
			{
				char *passwd ;
				do
				{
					passwd = getpass("WNC Password: ");
				}
				while ((!passwd) || (strlen(passwd) == 0)) ;
				if (strlen(passwd) > 8)
				{
					passwd[8] = '\0';
				}
				vncEncryptBytes(challenge, passwd);
				// "Forget" the clear-text password
				for (int i = strlen(passwd); i >= 0; i--)
				{
					passwd[i] = '\0';
				}
			}

			_conn->send((char *)challenge, CHALLENGESIZE) ;

			CARD32 authResult;
			RECEIVE((char *)&authResult, 4) ;
			authResult = Swap32IfLE(authResult);

			switch (authResult)
			{
			case rfbVncAuthOK:
#if DEBUG_LEVEL>=1
				std::cerr << "wncSource: WNC authentication "
					<< "succeeded" << std::endl ;
#endif
				break;
			case rfbVncAuthFailed:
				throw std::runtime_error(
					"wncSource: WNC authentication failed") ;
			case rfbVncAuthTooMany:
				throw std::runtime_error(
					"wncSource: WNC authentication "
					"failed - too many tries");
			default:
				throw std::runtime_error(
					"wncSource: Unknown WNC authentication "
					"result") ;
			}
		}
		break;

		default:
			throw std::runtime_error(
				"Unknown authentication scheme from WNC server");
		} /* switch (authScheme) */

		// -----------------------------------------------------------

		rfbClientInitMsg ci;
		ci.shared = 1 ;
		_conn->send((char *)&ci, sz_rfbClientInitMsg) ;

		rfbServerInitMsg si;
		RECEIVE((char *)&si, sz_rfbServerInitMsg) ;

		si.framebufferWidth = Swap16IfLE(si.framebufferWidth);
		si.framebufferHeight = Swap16IfLE(si.framebufferHeight);
		si.format.redMax = Swap16IfLE(si.format.redMax);
		si.format.greenMax = Swap16IfLE(si.format.greenMax);
		si.format.blueMax = Swap16IfLE(si.format.blueMax);
		si.nameLength = Swap32IfLE(si.nameLength);

		_width = si.framebufferWidth;
		_height = si.framebufferHeight;

		char *desktopName = new char [si.nameLength + 1] ;
		RECEIVE(desktopName, si.nameLength) ;

		desktopName[si.nameLength] = 0;

#if DEBUG_LEVEL>=1
		fprintf(stderr,"Desktop name \"%s\" (%ix%i)\n",desktopName,
			_width, _height);
		fprintf(stderr,"Connected to WNC server, using protocol "
			"version %d.%d\n",
			rfbProtocolMajorVersion, rfbProtocolMinorVersion);
		fprintf(stderr,"WNC server default format:\n");
		PrintPixelFormat(&si.format);
#endif
		delete [] desktopName;
		
		if (si.format.bitsPerPixel == 8 || si.format.bitsPerPixel == 16)
		{
			_depth = 16;
			_wncEncoding = Image::RGB565;
		}
		else
		{
			_depth = 24;
			_wncEncoding = Image::RGBA;
		}

		if (_texEncoding == Image::OPAQUE)
		{
			if (_depth == 16)
			{
				_texEncoding = Image::RGB565;
			}
			else
			{
				_texEncoding = Image::RGBA;
			}
		}

		// --- Pixel format ------------------------------------------

		rfbPixelFormat simple_format ;
#ifdef WORDS_BIGENDIAN
		simple_format.bigEndian = 1;
#else
		simple_format.bigEndian = 0;
#endif
		simple_format.trueColour = 1 ;
		if (_depth == 16)
		{
			simple_format.bitsPerPixel =  16;
			simple_format.depth = 16;
			simple_format.redMax = 31;
			simple_format.greenMax = 63;
			simple_format.blueMax = 31;

#ifndef WORDS_BIGENDIAN
			simple_format.redShift = 11;
			simple_format.greenShift = 5 ;
			simple_format.blueShift = 0 ;
#else
			simple_format.redShift = 0;
			simple_format.greenShift = 5;
			simple_format.blueShift = 11;
#endif
		}
		else // depth 24
		{
			simple_format.bitsPerPixel = 32 ;
			simple_format.depth = 24 ;
			simple_format.redMax =
				simple_format.greenMax =
				simple_format.blueMax = 255 ;
#ifndef WORDS_BIGENDIAN
			simple_format.redShift = 0 ;
			simple_format.greenShift = 8 ;
			simple_format.blueShift = 16 ;
#else
			simple_format.redShift = 24 ;
			simple_format.greenShift = 16 ;
			simple_format.blueShift = 8 ;
#endif
		}

		//simple_format.pad1 = 0 ;
		//simple_format.pad2 = 0 ;

		rfbSetPixelFormatMsg spf;
		spf.type = rfbSetPixelFormat;
		spf.format = simple_format ;
		spf.format.redMax = Swap16IfLE(spf.format.redMax);
		spf.format.greenMax = Swap16IfLE(spf.format.greenMax);
		spf.format.blueMax = Swap16IfLE(spf.format.blueMax);
		//spf.pad1 = 0;
		//spf.pad2 = 0;

		_conn->send((char *)&spf, sz_rfbSetPixelFormatMsg) ;

		// --- Encoding ----------------------------------------------


#if DEBUG_LEVEL>=1
		std::cerr << "Use " << Image::getEncodingName(_texEncoding)
			  << " texture and received "
			  << Image::getEncodingName(_wncEncoding)
			  << " images data"
			  << std::endl ;
#endif

		char buf[sz_rfbSetEncodingsMsg + 5 * 4]; // 5 = MAX_ENCODINGS
		rfbSetEncodingsMsg *se = (rfbSetEncodingsMsg *)buf;
		CARD32 *encs = (CARD32 *)(&buf[sz_rfbSetEncodingsMsg]);
		se->type = rfbSetEncodings;
		se->pad = 0;
		se->nEncodings = 0;

		// Not supported by Xwnc
		// encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCopyRect);
		// encs[se->nEncodings++] = Swap32IfLE(rfbEncodingHextile);
		// encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRRE);
		
		
		// broken here or in Xwnc when resizing a window ?
		//encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCoRRE);

		// any way raw encoding is the best encoding in the local
		// case ??

		if (_useShm)
		{
			encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRawShm);
#if DEBUG_LEVEL>=1
			std::cerr << "Use shared memory for framebuffer update"
				  << std::endl ;
#endif
		}
		else
		{
			encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);
		}


		encs[se->nEncodings++] = Swap32IfLE(rfbEncodingPointerPos);
		encs[se->nEncodings++] = Swap32IfLE(rfbEncodingXCursor);
#ifdef HAVE_XCURSOR
		encs[se->nEncodings++] = Swap32IfLE(rfbEncodingARGBCursor);
#endif
		encs[se->nEncodings++] = Swap32IfLE(rfbEncodingWindowShape);

		int len = sz_rfbSetEncodingsMsg + se->nEncodings * 4;
		se->nEncodings = Swap16IfLE(se->nEncodings);
		_conn->send(buf, len) ;

#if 0
		// --- Here we go ! ------------------------------------------
		_fk = FileKeeper::create(_conn->getFd(),  FileKeeper::R);
		subscribeTo(_fk) ;
		ReactiveEngine::notify(this, this) ;
#endif

		bitsPerSeconds(0, WNC_TIME_DEBUG_START, 0);

		return true ;
	} catch (std::runtime_error e) {
		std::cerr << e.what() << std::endl ;
		stop() ;
		return false ;
	}
}

void wncSource::startRequest(void)
{
	// --- Here we go ! ------------------------------------------
	_fk = FileKeeper::create(_conn->getFd(),  FileKeeper::R);
	subscribeTo(_fk) ;
	ReactiveEngine::notify(this, this) ;

	updateRequest(true);

	bitsPerSeconds(0, WNC_TIME_DEBUG_START, 0);
}

void
wncSource::react(Observable*)
{
	if (!_readWNCServer())
	{
		// error should exit
	}
	notifyObservers() ;
}

bool wncSource::_framebufferUpdate(rfbServerToClientMsg msg)
{
	RECEIVE(((char *)&msg.fu) + 1, sz_rfbFramebufferUpdateMsg - 1) ;
	unsigned long nRects = Swap16IfLE(msg.fu.nRects);
	unsigned char *shm = NULL, *s = NULL;
	unsigned long nshmRects = 0;
	unsigned long shmid = Swap32IfLE(msg.fu.shmid);
	unsigned long xWindow = Swap32IfLE(msg.fu.window);
	unsigned long isRoot = Swap32IfLE(msg.fu.topWindow);

	bitsPerSeconds(
		0, WNC_TIME_DEBUG_FBU_START, 0);

	if (isRoot == 0)
	{
		isRoot = 1;
	}
	else
	{
		isRoot = 0;
	}

#if DEBUG_LEVEL>=2
	fprintf(stderr, "wncSource[FramebufferUpdate]: window: 0x%lx (%i,%i)\n",
		xWindow, msg.fu.nRects, isRoot);
#endif

	XwncWindow *wncWin = 0;

	wncWin = _wncDesktop->findWindow(xWindow);
	float wWidth = 0, wHeight = 0;

	if (wncWin)
	{
		wncWin->getCurrentSize(&wWidth, &wHeight);
	}

	if (shmid)
	{
		shm = (unsigned char *)shmat(shmid, NULL, 0);
		if ((char *)shm == (char *) -1)
		{
			perror("shmat");
			shmid = 0;
		}
		else
		{
			s = shm;
			memcpy(&nshmRects, s, sizeof(CARD16));
			s += sizeof(CARD16);
			nshmRects = Swap16IfLE(nshmRects);
		}
	}


	int i;
	for (i=0; i< nRects+nshmRects; ++i)
	{
	  	unsigned char *conv = NULL;

		rfbFramebufferUpdateRectHeader rect;
		if (i < nRects)
		{
			RECEIVE((char *)&rect,
				sz_rfbFramebufferUpdateRectHeader) ;
		}
		else
		{
			memcpy(&rect, s, sz_rfbFramebufferUpdateRectHeader);
			conv = s + sz_rfbFramebufferUpdateRectHeader;
			s += sz_rfbFramebufferUpdateRectHeader;
		}

		rect.r.x = Swap16IfLE(rect.r.x);
		rect.r.y = Swap16IfLE(rect.r.y);
		rect.r.w = Swap16IfLE(rect.r.w);
		rect.r.h = Swap16IfLE(rect.r.h);
		rect.encoding = Swap32IfLE(rect.encoding);

#if 0
		std::cerr << "wncSource: rfbFramebufferUpdate " << rect.r.x
			  << "," << rect.r.y ;
		std::cerr << " " << rect.r.w << "x" << rect.r.h <<
			" encoding: " << rect.encoding << std::endl ;
#endif

		if (rect.encoding == rfbEncodingXCursor)
		{
			int bpr = (rect.r.w + 7) / 8;
			int bd = bpr * rect.r.h;
			char *buf = NULL;
			rfbXCursorColors colors;

			if (rect.r.w * rect.r.h)
			{
				RECEIVE((char *)&colors, sz_rfbXCursorColors);

				buf = (char *)malloc(bd * 2);
				if (buf == NULL)
				{
					std::cerr << "wncSource: malloc fail"
						  << std::endl;
					return False;
				}

				RECEIVE(buf, bd * 2);
			}
			_wncDesktop->HandleXCursor(
				rect.r.x, rect.r.y, rect.r.w, rect.r.h,
				buf, colors);
			continue;
		}

		if (rect.encoding == rfbEncodingARGBCursor)
		{
#ifdef HAVE_XCURSOR
			char *buf = NULL;
			unsigned int bufSize = rect.r.w * rect.r.h * 4;
			rfbXCursorColors colors;

			if (bufSize)
			{
				buf = (char *)malloc(bufSize);
				if (buf == NULL)
				{
					std::cerr << "wncSource: malloc fail"
						  << std::endl;
					return False;
				}

				RECEIVE((char *)buf, bufSize);
			}
			else
			{
				_wncDesktop->HandleXCursor(
					rect.r.x, rect.r.y, 0, 0,
					buf, colors);
			}
			_wncDesktop->HandleARGBCursor(
				(CARD32)rect.r.x, (CARD32)rect.r.y,
				(CARD32)rect.r.w, (CARD32)rect.r.h,
				(XcursorPixel *)buf);
#else // HAVE_XCURSOR
			// should not happen !
			unsigned	buffer_size = rect.r.w * rect.r.h * 4;
			char	buffer[buffer_size];
			RECEIVE(buffer, buffer_size);
#endif // HAVE_XCURSOR
			continue;
		}

		if (rect.encoding == rfbEncodingPointerPos)
		{
			_wncDesktop->handleCursorPosition(rect.r.x, rect.r.y);
			continue;
		}

		if (rect.encoding == rfbEncodingWindowShape)
		{
			/* rect.r contains the size of the window */
			rfbWindowShapeHeader h;
			RECEIVE((char *)&h, sz_rfbWindowShapeHeader) ;
			h.nrects = Swap32IfLE(h.nrects);
			int size =  h.nrects * 4 * sizeof(CARD16) ;
			CARD16 *buf = NULL;
			if (size)
			{
				buf = (CARD16 *)malloc(size);
				RECEIVE((char *)buf, size) ;
			}
			_wncDesktop->handleWindowShape(wncWin, buf, h.nrects);
			continue;
		}

		if ((rect.r.h * rect.r.w) == 0)
		{
			continue ;
		}
		
		if (wncWin &&
		    (rect.r.x+rect.r.w > (int)wWidth ||
		     rect.r.y+rect.r.h > (int)wHeight))
		{
			std::cerr << "wncSource [rfbFramebufferUpdate]: "
				  << "Rectangle too big ..." << std::endl ;
			std::cerr << "   " << rect.r.x << " " << rect.r.y
				  << " " << rect.r.w << "x" << rect.r.h
				  << " for: " << (int)wWidth << "x"
				  << (int)wHeight << std::endl ;
			// Fatal error?
			//continue ;
		}

		switch (rect.encoding)
		{
		case rfbEncodingRawShm:
		case rfbEncodingRaw:
		{
			int rectImgSize;

			rectImgSize = rect.r.w * rect.r.h *
				Image::getBytesPerPixel(_wncEncoding);

			bool isShm = false;

			Image tmpImg;
			tmpImg.setEncoding(_wncEncoding);

			if (shmid != 0)
			{
				isShm = true;
				tmpImg.setData(
					(unsigned char *)s, rectImgSize,
					Image::NONE);
				tmpImg.setDims(rect.r.w, rect.r.h);
				if (_texEncoding != _wncEncoding)
				{
					convertImage(&tmpImg, _texEncoding);
				}
				s += rectImgSize;
			}
			else
			{
				unsigned char *ptr =
					Image::AllocMem(rectImgSize) ;
				RECEIVE((char *)ptr,rectImgSize) ;
				tmpImg.setData(ptr, rectImgSize, Image::NONE) ;
				tmpImg.setDims(rect.r.w, rect.r.h) ;
				if (_texEncoding != _wncEncoding)
				{
					convertImage(&tmpImg, _texEncoding);
					free(ptr);
				}
			}
#if 0
			std::cerr << "raw: " << rect.r.w << "x" << rect.r.h <<
				" " << rect.r.x << "," << rect.r.y << std::endl ;
#endif
			_wncDesktop->handleImageFramebufferUpdate(
				wncWin, (bool)isRoot, &tmpImg,
				rect.r.x, rect.r.y, rect.r.w, rect.r.h);

			bitsPerSeconds(
				rectImgSize, WNC_TIME_DEBUG_BITS, isShm);
		}
		break; 

		default:
			std::cerr << "wncSource: unknown rect encoding:"
				  << rect.encoding << std::endl ;
			return false;
		}
	}

	if (shmid != 0)
	{
		shmdt(shm);
		if (shmctl(shmid, IPC_RMID, 0) != 0)
		{
			perror("shmctl IPC_RMID");
		}
	}

	bitsPerSeconds(
		0, WNC_TIME_DEBUG_FBU_STOP, 0);
	return true;
}

bool
wncSource::_readWNCServer(void)
{
	bool getAfbu = false;

	if (!_conn) return false;

	int fd = _conn->getFd() ;
	unsigned int i ;
	int count = 0;
	int fbcount = 0;
	int confCount = 0;
	int restackCount = 0;
	Chronometer _chrono;

	_chrono.start();

	while (_chrono.read() < 1000 && (getavail(fd) || _buffer.length()))
	  //while ((getavail(fd) || _buffer.length()))
	{
		rfbServerToClientMsg msg;
		RECEIVE((char *)&msg, 1) ;

		count++;

		switch (msg.type)
		{

		case rfbBell:
			std::cerr << "wncSource: ring my bell..." << std::endl ;
			break ;

		case rfbSetColourMapEntries:
		{
			RECEIVE(
				((char *)&msg.scme) + 1,
				sz_rfbSetColourMapEntriesMsg - 1) ;
			int size = msg.scme.nColours * 3 /* * CARD16 */ ;
			char *tmp = new char [size] ;
			RECEIVE(tmp, size) ;
			delete tmp ;
		}
		break ;

		case rfbServerCutText:
		{
			RECEIVE(((char *)&msg.sct) + 1,
				sz_rfbServerCutTextMsg - 1) ;
			msg.sct.length = Swap16IfLE(msg.sct.length);
			char *tmp = new char [msg.sct.length] ;
			RECEIVE(tmp, msg.sct.length) ;
			std::cerr << "wncSource: new text in cut buffer '"
				  << tmp << "'" << std::endl ;
			delete tmp;
		}
		break ;

		case rfbConfigureWindow:
		{
			RECEIVE(((char *)&msg.cw) + 1,
				sz_rfbConfigureWindowMsg - 1) ;
			msg.cw.window = Swap32IfLE(msg.cw.window);
			msg.cw.xsgn =  Swap16IfLE(msg.cw.xsgn);
			int x = Swap16IfLE(msg.cw.x);
			if (msg.cw.xsgn == 0)
			{
				x = -x;
			}
			msg.cw.ysgn =  Swap16IfLE(msg.cw.ysgn);
			int y = Swap16IfLE(msg.cw.y);
			if (msg.cw.ysgn == 0)
			{
				y = -y;
			}
			msg.cw.width = Swap32IfLE(msg.cw.width);
			msg.cw.height =	Swap32IfLE(msg.cw.height);
			_wncDesktop->handleConfigureWindow(
				msg.cw.window, x, y,
				(int)msg.cw.width, (int)msg.cw.height,
				Swap16IfLE(msg.cw.isroot));
			confCount++;
		}
		break;

		case rfbUnmapWindow:
		{
			RECEIVE(((char *)&msg.uw) + 1,
				sz_rfbUnmapWindowMsg - 1) ;
			msg.uw.window = Swap32IfLE(msg.uw.window);
			_wncDesktop->handleUnmapWindow(msg.uw.window);
		}
		break;

		case rfbDestroyWindow:
		{
			RECEIVE(((char *)&msg.dw) + 1,
				sz_rfbDestroyWindowMsg - 1) ;
			msg.dw.window = Swap32IfLE(msg.dw.window);
			_wncDesktop->handleDestroyWindow(msg.dw.window);
		}
		break;
    
		case rfbRestackWindow:
		{
			RECEIVE(((char *)&msg.rw) + 1,
				sz_rfbRestackWindowMsg - 1) ;
			msg.rw.window = Swap32IfLE(msg.rw.window);
			msg.rw.nextWindow = Swap32IfLE(msg.rw.nextWindow);
			msg.rw.transientFor = Swap32IfLE(msg.rw.transientFor);
			msg.rw.unmanagedFor = Swap32IfLE(msg.rw.unmanagedFor);
			msg.rw.flags = Swap32IfLE(msg.rw.flags);
			_wncDesktop->handleRestackWindow(
				msg.rw.window, msg.rw.nextWindow,
				msg.rw.transientFor, msg.rw.unmanagedFor,
				msg.rw.flags);
			restackCount++;
		}
		break;

		case rfbFramebufferUpdate:
		{
			bool ret = _framebufferUpdate(msg);
			if (!ret)
			{
				return false;
			}
			fbcount++;
			getAfbu = true;
		}
		break;

		default:
			std::cerr << "wncSource: unknown message type "
				  << (int)msg.type << " from WNC server"
				  << std::endl ;
			return false;
			break ;
		} /* switch */

	}

	if (getAfbu)
	{
		bitsPerSeconds(0, WNC_TIME_DEBUG_FRAME_STOP, 0);
	}

#if 0
	if (count >= 11 || _chrono.read() >= 100)
	{
	    std::cerr << "Chrono: " <<  _chrono.read() << " Count: " << count
		      << " FB: " << fbcount
		      << " Conf: " << confCount
		      << " Restack: " << restackCount << std::endl;
	}
#endif

	if (count)
	{
		updateRequest(true);
	}

	return true;
}

bool
wncSource::stop(void) {
  if (_conn) {
    delete _conn ;
    _conn=0 ;
    return true ;
  }
  return false ;
}

// ------------------------------------------------------------------

void
wncSource::updateRequest(bool incremental)
{
	updateRequest(0,0,10*_width,10*_height,incremental) ;
}

void
wncSource::updateRequest(int x, int y, int w, int h, bool incremental) {
  rfbFramebufferUpdateRequestMsg fur;

  fur.type = rfbFramebufferUpdateRequest;
  fur.incremental = incremental ? 1 : 0;
  fur.x = Swap16IfLE(x);
  fur.y = Swap16IfLE(y);
  fur.w = Swap16IfLE(w);
  fur.h = Swap16IfLE(h);

  _conn->send((char *)&fur, sz_rfbFramebufferUpdateRequestMsg) ;
}

void
wncSource::keyEvent(unsigned long key, bool down_flag) {
	rfbKeyEventMsg ke;

	ke.type = rfbKeyEvent ;
	ke.down = down_flag ? 1 : 0 ;

	ke.key = Swap32IfLE(key) ;
	_conn->send((char *)&ke, sz_rfbKeyEventMsg) ;
}

void
wncSource::pointerEvent(
	Window id, int x, int y, unsigned char button_mask)
{
	rfbPointerEventMsg pe;

#if 0
	std::cerr << "wncSource::pointerEvent: " << std::hex << id << std::dec
		  <<" " << x << " " << y << " "
		  << (int)button_mask << std::endl;
#endif

#if 0
	if (x < 0 || y < 0)
	{
		return;
	}
#endif

	pe.type = rfbPointerEvent ;
	pe.buttonMask = button_mask ;
	int xsgn = 1, ysgn = 1;

	if (x < 0)
	{
		xsgn = 0;
		x = -x;
	}
	if (y < 0) {
		ysgn = 0;
		y = -y;
	}
	pe.xsgn = Swap16IfLE(xsgn) ;
	pe.x = Swap16IfLE(x) ;
	pe.ysgn = Swap16IfLE(ysgn) ;
	pe.y = Swap16IfLE(y) ;
	pe.window = Swap32IfLE(id);
	_conn->send((char *)&pe, sz_rfbPointerEventMsg) ;
}

void
wncSource::pointerEvent(int x, int y, unsigned char button_mask)
{
	pointerEvent(0, x, y, button_mask);
}

void wncSource::getSize(unsigned int *width, unsigned int *height)
{
	*width = _width;
	*height = _height;
}
