#ifndef FoldableWindow_H
#define FoldableWindow_H

/*
 *
 * wstyle/FoldableWindow.H --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/core/TimeKeeper.H>
#include <nucleo/gl/window/keysym.H>

#include "AbstractWindowRenderer.H"
#include "FoldablePolygon.H"

class FoldableWindow : public AbstractWindowRenderer {

protected:

  FoldablePolygon _polygon ;
  bool _cropToCircle ;
  bool _testPolygon;
  TimeKeeper *_animTk; /* for animation */
  TimeKeeper *_foldTk; /* for folding back */
  
  bool _folding, _foldingBack ;
  double _x_from, _y_from, _x_to, _y_to ;

  glTiledTexturedImage::tiledTessellation _tiled_tessellation;
  bool _useTessCache;

  void reset(void) {
    double hw,hh ;
    _window->getSize(&hw,&hh) ;

    hw= hw/2.0;
    hh= hh/2.0;

    _useTessCache = false;
    _tiled_tessellation.clear();

    _polygon.points.clear() ;
    if (_cropToCircle) {
	 for (double a=0; a<2.0*M_PI; a+=0.1) {
	   double x=(hw-1.2)*cos(a), y=(hh-1.2)*sin(a) ;
	   _polygon.points.push_back(glTiledTexturedImage::point(x,y)) ;
	 }
    } else if (_testPolygon) {
	    GLdouble ch = 2*hh/3;
	    GLdouble cw = 2*hw/3;
	    _polygon.points.push_back(glTiledTexturedImage::point(-hw,-hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(0,-ch)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(hw,-hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(cw,0)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(hw,hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(0,ch)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(-hw,hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(-cw,0)) ;
    }
    else
    {
	    _polygon.points.push_back(glTiledTexturedImage::point(-hw,-hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(hw,-hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(hw,hh)) ;
	    _polygon.points.push_back(glTiledTexturedImage::point(-hw,hh)) ;
    }
    _polygon.reset() ;

    _folding = _foldingBack = false ;
    postRedisplay() ;
  }

  void _shapeDisplay(bool on)
  {
	  if (_nShapeRects == 0 || !_canUseStencilBuffer)
	  {
		  return;
	  }

	  if (on)
	  {
		  // this slow down things
		  //glClearStencil(0);
		  //glClear(GL_STENCIL_BUFFER_BIT);

		  glEnable(GL_STENCIL_TEST) ;

		  // test always fail     ref, mask
		  glStencilFunc(GL_NEVER, 0x1, 0x1);
		  // replace by ref if the test fail
		  glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);

		  float hw,hh;
		  _window->getSize(&hw, &hh) ;
		  hw=hw/2.0; hh=hh/2.0 ;
		  int i;
		  for (i = 0; i < _nShapeRects; i++)
		  {
			  float x = _shape[i*4];
			  float y = _shape[i*4 + 1];
			  float w = _shape[i*4 + 2];
			  float h = _shape[i*4 + 3];

			  GLdouble x1, y1, x2, y2;
			  x1 = -hw + x;
			  y1 = hh - y;
			  x2 = x1 + w;
			  y2 = y1 - h;
			  glRectd(x1,y1,x2,y2);
		  }

		  glStencilFunc(GL_EQUAL, 0x1, 0x1);
		  glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
	  }
	  else
	  {
		  glDisable(GL_STENCIL_TEST);
	  }
  }

  void _shapeSelect(bool on) {
	  /* Select does not take in accoun the STENCIL buffer. So, there is:
	   *    AbstractWindowRenderer::isInsideWindow and
	   * and
	   *    AScreen::_myPickClosest
	   * to take in account the shape */
  }

  void display(dlPolicy policy) {
    float width,height;
    _window->getSize(&width,&height) ;

    float red, green, blue, alpha ;
    _window->getBgcolor(&red, &green, &blue, &alpha) ;

    std::list<glTiledTexturedImage::point>::iterator i ;

    // ----------------------------------

    _shapeDisplay(true);

     // --------------------------------

    glTiledTexturedImage *texture = 0 ;

    int texCoordsSize = 0 ;
    float *texCoords = 0 ;
    _window->getTextureMapping(&texture, &texCoordsSize, &texCoords) ;

    glColor4f(red,green,blue,alpha);

    if (texture && texCoordsSize==8 && texCoords) {
	 float left = texCoords[0]  ;
	 float bottom = texCoords[1] ;
	 float right = texCoords[2] ;
	 float top = texCoords[5] ;
	 float Ox = left+(right-left)/2.0 ;
	 float Oy = top+(bottom-top)/2.0 ;

	 texture->enable() ;

	 if (_testPolygon || _folding || _foldingBack || _cropToCircle)
	 {
		 // tessellator should be used and are used if the nbr of 
		 // texture in the tile is > 1
		 texture->tessDisplay(&(_polygon.poly1), _useTessCache);
		 _useTessCache = true;
		 texture->getLastTessellation(&_tiled_tessellation);
	 }
	 else
	 {
		 _useTessCache = false;
		 texture->display();
	 }

	 texture->disable() ;

	 if (_nShapeRects == 0)
	 {
		 glDisable(GL_DEPTH_TEST) ;
		 // Hum ...
		 glColor4f(0.5,0.5,0.5,0) ;
		 glBegin(GL_LINE_LOOP) ;
		 for (i=_polygon.poly1.begin(); i!=_polygon.poly1.end(); ++i) {
			 double xp=(*i).x, yp=(*i).y ;
			 double xt = Ox + xp*(right-left)/width ;
			 double yt = Oy + yp*(top-bottom)/height ;
			 //glTexCoord2d(xt, yt) ;
			 glVertex2d(xp,yp) ;
		 }
		 glEnd() ;
		 glEnable(GL_DEPTH_TEST) ;
	 }

    } else {
	 glBegin(GL_POLYGON) ;
	 for (i=_polygon.poly1.begin(); i!=_polygon.poly1.end(); ++i) {
	   glVertex2d((*i).x,(*i).y) ;
	 }
	 glEnd() ;
    }

    if (_polygon.poly2.size()) {
	 glTranslatef(0,0,1) ;
	 glColor4f(0.6*red,0.6*green,0.6*blue,alpha*0.5) ;
	 glBegin(GL_POLYGON) ;
	 for (i=_polygon.poly2.begin(); i!=_polygon.poly2.end(); ++i) {
	   glVertex2d((*i).x, (*i).y) ;
	 }
	 glEnd() ;
    }
    
    /* add shadow */
    if ((_shadow && !_polygon.poly2.size() && !getWindow()->isUnmanaged() &&
	 !_cropToCircle && !getWindow()->isRootWindow() && _nShapeRects == 0)
	|| (getWindow()->isToolTips()))
    {
      float hw, hh;
      _window->getSize(&hw, &hh) ;
      hw=(int)(hw/2.0); hh=(int)(hh/2.0);

      float startAlpha = 0.3 * getWindow()->getAlpha();

      GLboolean isBlendEnabled ;
      glGetBooleanv(GL_BLEND, &isBlendEnabled) ;
      if (!isBlendEnabled) glEnable(GL_BLEND) ;

#if 0
      GLfloat shborder = 5 ;
      glColor4f(0.1,0.1,0.1, startAlpha) ;
      glPushMatrix() ;
      glTranslatef(0,0,-0.05) ;
      glRectf(-hw+shborder,-hh-shborder,
	      hw+shborder,hh-shborder) ;
      glPopMatrix() ;
#endif

      GLfloat shborder = 8 ;

      if (getWindow()->isToolTips()) {
	shborder = 6 ;
      }

#define LEFTRIGHT_SHADOW

#ifdef LEFTRIGHT_SHADOW
      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(-hw+shborder, -hh);
      glVertex2f(hw, -hh);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(hw, -hh-shborder);
      glVertex2f(-hw+shborder, -hh-shborder);
      glEnd() ;

      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(hw, -hh);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(hw+shborder, -hh);
      glVertex2f(hw+shborder, -hh-shborder);
      glVertex2f(hw, -hh-shborder);
      glEnd() ;

      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(hw, -hh);
      glVertex2f(hw, hh-shborder);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(hw+shborder, hh-shborder);
      glVertex2f(hw+shborder, -hh);
      glEnd() ;
#else
      // top
      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(-hw, hh);
      glVertex2f(hw, hh);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(+hw+shborder, hh+shborder);
      glVertex2f(-hw-shborder, hh+shborder);
      glEnd() ;

      // bottom
      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(-hw, -hh);
      glVertex2f(hw, -hh);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(hw+shborder, -hh-shborder);
      glVertex2f(-hw-shborder, -hh-shborder);
      glEnd() ;

      // right
      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(hw, -hh);
      glVertex2f(hw, hh);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(hw+shborder, hh+shborder);
      glVertex2f(hw+shborder, -hh-shborder);
      glEnd() ; 

      // left 
      glBegin(GL_POLYGON) ;
      glColor4f(0.1,0.1,0.1,startAlpha) ;
      glVertex2f(-hw, hh);
      glVertex2f(-hw, -hh);
      glColor4f(0.1,0.1,0.1,0) ;
      glVertex2f(-hw-shborder, -hh-shborder);
      glVertex2f(-hw-shborder, +hh+shborder);
      glEnd() ; 
#endif
      if (!isBlendEnabled) glDisable(GL_BLEND) ;
    }

    _shapeDisplay(false);

    if (texCoords) {
	    delete [] texCoords;
    }
  }

  void select(void)
  {
    std::list<glTiledTexturedImage::point>::iterator i ;
    bool folded = _polygon.poly2.size()>0 ;

    if (getWindow() != 0 && getWindow()->isToolTips())
    {
	    return;
    }

    if (getWindow() != 0 && getWindow()->doPseudoSelection())
    {
	    pseudoSelect();
	    return;
    }

    _shapeSelect(true);

    if (_useTessCache && _tiled_tessellation.size() > 0)
    {
	    glTiledTexturedImage::tiledTessellation:: iterator itt;
	    glTiledTexturedImage::tessellation:: iterator it;
	    std::list<glTiledTexturedImage::point >:: iterator ip;

	    for (itt = _tiled_tessellation.begin();
		 itt != _tiled_tessellation.end(); ++itt)
	    {
		    for (it = (*itt).begin(); it != (*itt).end(); ++it)
		    {
			    glBegin((*it).type);
			    for (ip = (*it).points.begin();
				 ip != (*it).points.end(); ++ip)
			    {
				    glVertex2d((*ip).x, (*ip).y);
			    }
			    glEnd();     
		    }
	    }
	    
    }
    else
    {
	    glBegin(GL_POLYGON) ;
	    for (i=_polygon.poly1.begin(); i!=_polygon.poly1.end(); ++i) {
		    float xx = (*i).x;
		    float yy = (*i).y;
#if 0
		    // FIXME this fix unproject but this is not really correct
		    //if (xx < 0) xx += 2;
		    if (xx > 0) xx -= 1;
		    //if (yy < 0) yy += 2;
		    if (yy > 0) yy -= 1;
		    glVertex2d(xx,yy) ;
#else
		    glVertex2d(xx,yy) ;
#endif

	    }
	    glEnd() ;
    }

    if (folded) {
	 glTranslatef(0,0,1) ;
	 glBegin(GL_POLYGON) ;
	 for (i=_polygon.poly2.begin(); i!=_polygon.poly2.end(); ++i) {
	   glVertex2d((*i).x, (*i).y) ;
	 }
	 glEnd() ;
    }

    _shapeSelect(false);
    // sgNode::select() ;
  }


  void pseudoSelect()
  {
	  double hw,hh ;
	  _window->getSize(&hw,&hh) ;

	  hw= (hw/2.0)*10000;
	  hh= (hh/2.0)*10000;

	  glBegin(GL_POLYGON) ;
	  glVertex2d(-hw, -hh);
	  glVertex2d(hw, -hh);
	  glVertex2d(hw, hh);
	  glVertex2d(-hw, hh);
	  glEnd();
  }

void react(Observable *obs)
{
	if (_foldingBack && obs==_foldTk)
	{
		double dx = (_x_to-_x_from) ;
		double dy = (_y_to-_y_from) ;
		double d = dx*dx+dy*dy ;

		_useTessCache = false;
		_x_to -= dx*0.2 ;
		_y_to -= dy*0.2 ;
		int timer = (int)(d/1000.0) ;
		if (!timer)
		{
			_foldingBack = false ;
			unsubscribeFrom(_foldTk) ;
			delete _foldTk ;
			_foldTk = 0 ;
			reset() ;
		}
		else
		{
			_polygon.fold(_x_from,_y_from,_x_to,_y_to) ;
			_foldTk->arm(50);
			postRedisplay() ;
			
		}
	}

	if (obs==_animTk)
	{
		rotate_rel(-1,0,0,1) ;
		_animTk->arm(20);
	}
}

public:

  FoldableWindow(XwncWindow *window) :
    AbstractWindowRenderer(window,"FoldableWindow")
  {
    _cropToCircle = false ;
    _testPolygon = false;
    _animTk = 0;
    _foldTk = 0;
     reset() ;
  }

  ~FoldableWindow(void) {
    if (_animTk)
      delete _animTk ;
    if (_foldTk)
      delete _foldTk ;
  }


  void simpleReset(void) {
    reset();
  }

  void fold(int px, int py, bool fold) {
    if (fold) {
      if (_folding) {
	_x_to = px ;
	_y_to = py ;
	_polygon.fold(_x_from, _y_from, _x_to, _y_to) ; 
	postRedisplay() ;
      } else {
	_x_from = px ;
	_y_from = py ;
	_folding = true ;
      }
    } else {
      if (_folding) _foldingBack = true ;
      _folding = false ;
      _foldTk = TimeKeeper::create() ;
      subscribeTo(_foldTk);
      _foldTk->arm(50);
    }
    _useTessCache = false;
  }

  void cropToCircule(void) {
    _cropToCircle = !_cropToCircle ;
    _testPolygon = false;
    _useTessCache = false;
    reset() ;
  }

  void testPolygon(void) {
    _testPolygon = !_testPolygon;
    _cropToCircle = false;
    _useTessCache = false;
    reset() ;
  }

  void animate(void) {
    if (_animTk) {
	 unsubscribeFrom(_animTk) ;
	 delete _animTk ;
	 _animTk = 0 ;
    } else {
      _animTk = TimeKeeper::create(1) ;
      _animTk->arm(1);
      subscribeTo(_animTk) ;
    }
  }

} ;

#endif
